# -*- coding: UTF-8 -*-
#/**
# * Software Name : libmich 
# * Version : 0.2.2
# *
# * Copyright © 2012. Benoit Michau.
# * Made in TGV Paris Grenoble
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License version 2 as published
# * by the Free Software Foundation. 
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# * GNU General Public License for more details. 
# *
# * You will find a copy of the terms and conditions of the GNU General Public
# * License version 2 in the "license.txt" file or
# * see http://www.gnu.org/licenses/ or write to the Free Software Foundation,
# * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
# *
# *--------------------------------------------------------
# * File Name : core/shtr.py
# * Created : 2012-04-09
# * Authors : Benoit Michau 
# *--------------------------------------------------------
#*/ 

from struct import pack, unpack

# export filter
__all__ = ['decomposer', 'decompose', 'shtr']

# Utility classes to :
# decompose and recompose integers
# and manipulate str like integers (shifting them)

class decomposer(object):
    '''
    little utility to decompose an integer into factors from an 
    exponential (self.MUL), and returns the list of remainders:
    
    example:
    decomposer(0x10).decompose(53) -> [5, 3]
    this means:
    53 = 5 + (3 << 4), with 0x10 = 1 << 4
    
    more generally:
    decomposer(1<<X).decompose(Y) -> [a, b, c, d ...]
    means:
    Y = a + (b << (1<<X)) + (c << (2<<X)) + (d << (3<<X)) ...
    '''
    def __init__(self, MUL=0x100):
        self.MUL = MUL
        self._val_dec = []
    
    def decompose(self, val):
        self._val_dec.append( int(val % self.MUL) )
        if val >= self.MUL:
            val = val // self.MUL
            void = self.decompose( val )
        return self._val_dec

def decompose(MUL=0x100, val=0):
    '''
    little utility to decompose an integer into factors from an 
    exponential (MUL), and returns the list of remainders:
    
    example:
    decompose(0x10, 53) -> [5, 3]
    this means:
    53 = 5 + (3 << 4), with 0x10 = 1 << 4
    
    more generally:
    decomposer(1<<X, Y) -> [a, b, c, d ...]
    means:
    Y = a + (b << (1<<X)) + (c << (2<<X)) + (d << (3<<X)) ...
    
    This is faster than the recursive decomposer().decompose() method
    '''
    if MUL == 2:
        dec = map(int, bin(val)[2:])
        dec.reverse()
        return dec
    elif MUL == 16 and val >= 10000000000:
        h = hex(val)[2:]
        if h[-1] == 'L': h = h[:-1]
        dec = map(int, h, len(h)*[16])
        dec.reverse()
        return dec
    else:
        dec = [ int(val % MUL) ]
        while val >= MUL:
            val //= MUL
            dec.append( int(val % MUL) )
        return dec

class shtr(str):
    '''
    shtr are strings that can be shifted !
    
    When a shtr is shifted (<< X or >> X), it returns the resulting string.
    When right shifted (>> X), null bytes are appended left-side.
    
    Methods .left_val(X) and .right_val(X) returns the integer value corresponding to 
    the X bits left or right side of the shtr.
    '''
    
    def __init__(self, s):
        self._bitlen = len(s)*8
    
    def left_val(self, bitlen):
        '''
        returns big endian integer value from the `bitlen' left bits of the shtr
        '''
        if bitlen > self._bitlen:
            bitlen = self._bitlen
        # 0)
        reg64 =  bitlen//64
        reg8  = (bitlen%64)//8
        reg1  =  bitlen%8
        acc, start, stop = 0, 0, 0
        # 1)
        if reg64:
            stop = 8*reg64
            acc += reduce(lambda x,y: (x<<64)+y,
                          unpack('>'+reg64*'Q', self[:stop]))
        # 2)
        if reg8:
            acc <<= 8*reg8
            start = stop
            stop = start + reg8
            acc += reduce(lambda x,y: (x<<8)+y,
                          unpack('>'+reg8*'B', self[start:stop]))
        # 3)
        if reg1:
            acc <<= reg1
            acc += (ord(self[stop:stop+1]) >> (8-reg1))
        #
        return acc
    
    def right_val(self, val):
        '''
        returns big endian integer value from the `val' right bits of the shtr
        '''
        if val > self._bitlen: val = self._bitlen
        acc = 0
        # 1) get value of full bytes
        for i in xrange(val/8):
            acc += ord(self[-1-i]) << (i*8)
        # 2) get value of last bits
        if val%8 and val/8 < len(self):
            acc += (ord(self[-1-(val/8)]) & ((1<<(val%8))-1)) << (8*(val/8))
        return acc
    
    def __lshift__(self, bitlen):
        '''
        returns resulting shtr after shifting left of `bitlen' bits
        '''
        if bitlen > self._bitlen:
            bitlen = self._bitlen
        # 0) full byte shifting
        buf = self[bitlen//8:]
        # 1) bit shifting
        reg1  = bitlen%8
        if reg1:
            buflen = len(buf)
            reg64 = buflen//8
            reg8  = buflen%8
            reg1inv   = 8-reg1
            reg1inv64 = 64-reg1
            fmt = '>'+reg64*'Q'+reg8*'B'
            values = unpack(fmt, buf)
            #print values
            if reg64:
                #print('reg64')
                chars = map(lambda X: ((X[0]<<reg1)%0x10000000000000000)\
                                      +(X[1]>>reg1inv64),
                            zip(values[:reg64], values[1:reg64]))
                if reg8:
                    #print('reg8')
                    chars.append( ((values[reg64-1]<<reg1)%0x10000000000000000)\
                                  +(values[reg64]>>reg1inv) )
                    chars.extend( map(lambda X: ((X[0]<<reg1)%0x100)+(X[1]>>reg1inv),
                                      zip(values[reg64:], values[reg64+1:])) )
                    chars.append( (values[-1]<<reg1)%0x100 )
                else:
                    chars.append( (values[reg64-1]<<reg1)%0x10000000000000000 )
            elif reg8:
                #print('reg8')
                chars = map(lambda X: ((X[0]<<reg1)%0x100)+(X[1]>>reg1inv),
                            zip(values, values[1:]))
                chars.append( (values[-1]<<reg1)%0x100 )
            #
            #print chars
            ret = shtr(pack(fmt, *chars))
        else:
            ret = shtr(buf)
        ret._bitlen = self._bitlen-bitlen
        return ret
    
    def __rshift__(self, val):
        '''
        returns resulting shtr after shifting right of `val' bits
        '''
        # 1) handle full byte shifting
        Bsh = val / 8
        buf = ''.join((Bsh*'\0', self[:len(self)-Bsh]))
        # 2) then bit shifting
        bsh = val % 8
        pbsh, invbsh = pow(2,bsh)-1, 8-bsh
        # go over the string, from last byte to the 1st one
        strlist = [ (ord(buf[i])>>bsh) + ((ord(buf[i-1])&pbsh)<<invbsh) \
                    for i in reversed(xrange(1, len(buf))) ]
        # add 1st bits of the string and reverse the list
        strlist.append( ord(buf[0])>>bsh )
        strlist.reverse()
        #
        #return shtr(''.join(map(chr, strlist)))
        ret = shtr(''.join(map(chr, strlist)))
        ret._bitlen = self._bitlen + val
        return ret
#
#


# −*− coding: UTF−8 −*−
#/**
# * Software Name : libmich 
# * Version : 0.3.0
# *
# * Copyright © 2012. Benoit Michau. France Telecom.
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License version 2 as published
# * by the Free Software Foundation. 
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# * GNU General Public License for more details. 
# *
# * You will find a copy of the terms and conditions of the GNU General Public
# * License version 2 in the "license.txt" file or
# * see http://www.gnu.org/licenses/ or write to the Free Software Foundation,
# * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
# *
# *--------------------------------------------------------
# * File Name : core/element.py
# * Created : 2011-08-28 
# * Authors : Benoit Michau 
# *--------------------------------------------------------
#*/ 

#------------------------------------------------------------------------------#
#                                   The Libmich                                #
#------------------------------------------------------------------------------#
# Python library, works with Python 2.6 and 2.7
# not tested with older version
# and not going to support Python 3 (str / bytes is PITA here)
#
# version: 0.3
# author: Benoit Michau
#
# defines 3 kind of primary elements:
#   `Str': for byte stream
#   `Bit': for bit stream (actually managed like an integer value)
#   `Int': for integer value
#   all are subclasses of the common `Element' class
#
# elements can be stacked into a layer: 
#   `Layer': stacks [Str, Bit, Byte, Layer] (yes, it's recursive)
#   allows to manage dependencies between elements within the layer,
#   and with surrounding layers (next, previous, header, payload)
#   when placed in a block
#   
# layers can be stacked into a block:
#   `Block': stacks [Layer] (and only Layer)
#   allows to manage intelligently dependencies between layers 
#   (next, previous, header, payload)
#   with a hierarchy attribute assigned to each layer.
#
# Particularly convinient for building / parsing complex and mixed data 
# structure like network protocols: IKEv2, SCTP, Diameter, EAP, UMA...
# or like file structure: zip, PNG, ELF, MPEG4, ...
#
# TODO:
# + actually, still a lot !
# + defines plenty of formats
#------------------------------------------------------------------------------#

#------------------------------------------------------------------------------#
# import business
#------------------------------------------------------------------------------#
# check python version for deepcopy bug in 2.6
# and raise if Python version is not correct
import sys
def __version_err():
    print('[ERR] only python 2.6 and 2.7 are supported (unfortunately)')
    raise(Exception)
if sys.version_info[0] == 2:
    if sys.version_info[1] == 6:
        import copy
        import types
        def _deepcopy_method(x, memo):
            return type(x)(x.im_func, deepcopy(x.im_self, memo), x.im_class)
        copy._deepcopy_dispatch[types.MethodType] = _deepcopy_method
    elif sys.version_info[1] != 7:
        __version_err()
else:
    __version_err()

# export filtering
#
# exports the following constants:
# type_funcs, debug_level, ERR, WNG, DBG,
# functions: 
# log(), debug(), show(), showattr(), test()
# classes:
# Element(), Str(), Int(), Bit(), Layer(), RawLayer(), Block()
__all__ = ['Element', 'Str', 'Int', 'Bit', 'Layer', 'RawLayer', 'Block',
           'type_funcs', 'debug_level', 'ERR', 'WNG', 'DBG', 'log', 'debug',
           'show', 'showattr',
           'test', 'test0', 'test1', 'test2',
           'testTLV', 'testA', 'testB']

from copy import deepcopy
from struct import pack, unpack
from socket import inet_ntoa
from binascii import hexlify, unhexlify
from re import split, sub

from hydratk.extensions.datagen.asn1.core.shar import shar
from hydratk.extensions.datagen.asn1.core.shtr import shtr, decomposer, decompose
# TODO: cleanup the rest of libmich-related code to remove this import here
from hydratk.extensions.datagen.asn1.utils.repr import show, showattr

#------------------------------------------------------------------------------#
# helping routines
#------------------------------------------------------------------------------#
# defines a tuple of function-like 
class Dummy(object):
    def __init__(self):
        pass
type_funcs = ( type(lambda x:1), \
               type(Dummy().__init__), \
               type(inet_ntoa) )
del Dummy
#
# defines debugging facility
debug_level = {1:'ERR', 2:'WNG', 3:'DBG'}
ERR = 1
WNG = 2
DBG = 3

def log(level=DBG, string=''):
    # if needed, can be changed to write somewhere else
    # will redirect all logs from the library
    print('[libmich %s] %s' % (debug_level[level], string))

def debug(thres, level, string):
    if level and level<=thres:
        print('[%s] %s' %(debug_level[level], string))

#------------------------------------------------------------------------------#
# basic Element definitions
#------------------------------------------------------------------------------#
# Now defines Elements: Str, Int, Bit
# Element is a wrapping object for all the 3 following
# and has some common methods 
class Element(object):
    '''
    encapsulating class for: Str, Bit, Int
    '''
    # checking Element boundaries extensively
    #safe = True
    safe = False
    
    # Element debugging threshold: 0, ERR, WNG, DBG
    dbg = ERR
    
    # value assignment facilities
    def __lt__(self, Val):
        self.Val = Val
    
    def __gt__(self, Pt):
        self.Pt = Pt
    
    # transparency handling, common to all Element:
    def is_transparent(self):
        if self.TransFunc is not None:
            if self.safe: 
                assert( type(self.TransFunc(self.Trans)) is bool )
            if self.TransFunc(self.Trans): 
                return True
            else:
                return False
        elif self.Trans:
            if self.safe:
                assert(type(self.Trans) is bool)
            return True
        else:
            return False
    
    # an Element can point to another Element or Layer
    def getobj(self):
        if isinstance(self.Pt, (Element, Layer, tuple, list)) and not self.Val:
            return self.Pt
        else:
            return self()
    
    # this is to get a nice object representation:
    # can possibly be called with `show(element)`
    def show(self, with_trans=False):
        tr, re = '', ''
        if self.is_transparent():
            # TODO: eval the best convinience here
            if not with_trans:
                return ''
            tr = ' - transparent'
        else:
            tr = ''
        if self.ReprName != '':
            re = ''.join((self.ReprName, ' '))
        return '<%s[%s%s] : %s>' % ( re, self.CallName, tr, repr(self) )
    
    # when willing to bit shift str, call this instead of standard __str__()
    def shtr(self):
        return shtr(self.__str__())
    
    # this is to retrieve element's dynamicity from a mapped buffer
    def reautomatize(self):
        if self.Val is not None:
            if not self.PtFunc:
                self.Pt = self.Val
            self.Val = None
    
	# this is for uniformity with Block()
    def parse(self, s=''):
        self.map(s)

class Str(Element):
    '''
    class defining a standard Element, 
    managed like a stream of byte(s), or string.
    It is always byte-aligned (in term of length, at least)
    
    attributes:
    Pt: to point to another stream object (can simply be a string);
    PtFunc: when defined, PtFunc(Pt) is used 
            to generate the str() / len() representation;
    Val: when defined, overwrites the Pt (and PtFunc) string value, 
         used when mapping a string buffer to the element;
    Len: can be set to a fixed int value, or to another object
         when called by LenFunc
    LenFunc: to be used when mapping string buffer with variable length
             (e.g. in TLV object), LenFunc(Len) is used;
    Repr: python representation; binary, hexa, human or ipv4;
    Trans: to define transparent element which has empty str() and len() to 0,
           it "nullifies" its existence; can point to something for automation;
    TransFunc: when defined, TransFunc(Trans) is used to automate the 
               transparency aspect: used e.g. for conditional element;
    '''
    
    # this is used when printing the object representation
    _repr_limit = 1024
    _reprs = ['hex', 'bin', 'hum', 'ipv4']
    
    # padding is used when .Pt and .Val are None, 
    # but Str instance has still a defined .Len attribute
    _padding_byte = '\0'
    
    def __init__(self, CallName='', ReprName=None, 
                 Pt=None, PtFunc=None, Val=None, 
                 Len=None, LenFunc=None,
                 Repr="hum",
                 Trans=False, TransFunc=None):
        if CallName or not self.CallName:
            self.CallName = CallName
        if ReprName is None :
            self.ReprName = ''
        else :
            self.ReprName = ReprName
        self.Pt = Pt
        self.PtFunc = PtFunc
        self.Val = Val
        self.Len = Len
        self.LenFunc = LenFunc
        self.Type = 'stream'
        self.Repr = Repr
        self.Trans = Trans
        self.TransFunc = TransFunc
    
    def __setattr__(self, attr, val):
        # ensures no bullshit is provided into element's attributes 
        # (however, it is not a exhaustive test...)
        # managed with the class "safe" trigger
        if self.safe :
            if attr == 'CallName' :
                if type(val) is not str or len(val) == 0 :
                    raise AttributeError('CallName must be a non-null string')
            elif attr == 'ReprName' :
                if type(val) is not str:
                    raise AttributeError('ReprName must be a string')
            elif attr == 'PtFunc' :
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('PtFunc must be a function')
            elif attr == 'Val' :
                if val is not None and not isinstance(val, \
                (str, bytes, Element, Layer, Block, tuple, list)) :
                    raise AttributeError('Val must be a string or something ' \
                    'that makes a string at the end...')
            elif attr == 'Len' :
                if val is not None and not isinstance(val, \
                (int, tuple, Element, type_funcs)) :
                    raise AttributeError('Len must be an int or element')
            elif attr == 'LenFunc' :
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('LenFunc must be a function')
            elif attr == 'Repr' :
                if val not in self._reprs :
                    raise AttributeError('Repr %s does not exist, only: %s' \
                          % (val, self._reprs))
            elif attr == 'TransFunc' :
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('TransFunc must be a function')
        # this is for Layer() pointed by Pt attr in Str() object
        #if isinstance(self.Pt, Layer) and hasattr(self.Pt, attr):
        #    setattr(self.Pt, attr, val)
        # ...does not work properly
        # and the final standard python behaviour
        object.__setattr__(self, attr, val)
    
    def __getattr__(self, attr):
        # this is for Layer() pointed by Pt attr in Str() object
        if isinstance(self.Pt, Layer) and hasattr(self.Pt, attr):
            return getattr(self.Pt, attr)
        # and the final standard python behaviour
        object.__getattr__(self, attr)
    
    # the libmich internal instances check
    # this is highly experimental...
    def __is_intern_inst(self, obj):
        return isinstance(obj, (Element, Layer, Block))
    
    # building basic methods for manipulating easily the Element 
    # from its attributes
    def __call__(self, l=None):
        # when length has fixed value:
        if not l and type(self.Len) is int:
            l = self.Len
        #else:
        #    l = None
        # when no values are defined at all:
        if self.Val is None and self.Pt is None: 
            if l: return l * self._padding_byte
            else: return ''
        # returning the right string:
        # if defined, self.Val overrides self.Pt capabilities
        elif self.Val is not None:
            # allow to pass tuple or list of libmich internal instances
            if isinstance(self.Val, (list, tuple)) \
            and False not in map(self.__is_intern_inst, self.Val):
                return ''.join(map(str, self.Val))[:l]
            return str(self.Val)[:l]
        # else: use self.Pt capabilities to get the string
        elif self.PtFunc is not None: 
            if self.safe: 
                assert(hasattr(self.PtFunc(self.Pt), '__str__'))
            return str(self.PtFunc(self.Pt))[:l]
        else:
            # allow to pass tuple or list of libmich internal instances
            if isinstance(self.Pt, (list, tuple)) \
            and False not in map(self.__is_intern_inst, self.Pt):
                return ''.join(map(str, self.Pt))[:l]
            # otherwise, handle simply as is
            if self.safe: 
                assert(hasattr(self.Pt, '__str__'))
            return str(self.Pt)[:l]
    
    def __str__(self):
        # when Element is Transparent:
        if self.is_transparent():
            return ''
        #else:
        return self()
    
    def __len__(self):
        # does not take the LenFunc(Len) into account
        # When a Str is defined, the length is considered dependent of the Str
        # the Str is dependent of the LenFunc(Len) only 
        # when mapping data into the Element
        return len(self.__str__())
    
    def bit_len(self):
        return len(self)*8
    
    def map_len(self):
        # need special length definition 
        # when mapping a string to the 'Str' element 
        # that has no fixed length
        #
        # uses LenFunc, applied to Len, when length is variable:
        # and TransFunc, applied to Trans, to managed potential transparency
        # (e.g. for optional element triggered by other element)
        #
        if self.Len is None:
            return None
            #return 0
        if self.LenFunc is None: 
            return self.Len
        else:
            if self.safe:
                assert( type(self.LenFunc(self.Len)) in (int, long) )
            return self.LenFunc(self.Len)
    
    def __int__(self):
        # big endian integer representation of the string buffer
        return shtr(self).left_val(len(self)*8)
    
    def __bin__(self):
        # does not use the standard python 'bin' function to keep 
        # the right number of prefixed 0 bits
        h = hex(self)
        binary = ''
        for i in xrange(0, len(h), 2):
            b = format( int(h[i:i+2], 16), 'b' )
            binary += ( 8-len(b) ) * '0' + b
        return binary
    
    def __hex__(self):
        return self().encode('hex')
    
    def __repr__(self):
        # check for simple representations
        if self.Pt is None and self.Val is None: 
            return repr(None)
        if self.Repr == 'ipv4':
            #if self.safe: assert( len(self) == 4 )
            if len(self) != 4:
                return '0x%s' % hex(self)
            return inet_ntoa( self.__str__() )
        elif self.Repr == 'hex': 
            ret = '0x%s' % hex(self)
        elif self.Repr == 'bin': 
            ret = '0b%s' % self.__bin__()
        # check for the best human-readable representation
        elif self.Repr == 'hum':
            # standard return
            ret = repr( self() )
            # complex return:
            # allow to assign a full Block or Layer to a Str...
            if self.__is_intern_inst(self.Pt):
                ret = repr(self.Pt)
            if self.__is_intern_inst(self.Val):
                ret = repr(self.Val)
            # allow to assign a list or tuple of Block or Layer...
            if isinstance(self.Pt, (list, tuple)) \
            and False not in map(self.__is_intern_inst, self.Pt):
                ret = '|'.join(map(repr, self.Pt))
            if isinstance(self.Val, (list, tuple)) \
            and False not in map(self.__is_intern_inst, self.Val):
                ret = '|'.join(map(repr, self.Val))
            # finally, self.Val can be a raw value... still
            if self.Val is not None and hasattr(self.Val, '__repr__'):
                ret = repr(self.Val)
        # truncate representation if string too long:
        # avoid terminal panic...
        if len(ret) <= self._repr_limit:
            return ret
        else:
            return ret[:self._repr_limit-3]+'...'
    
    # some more methods for checking Element's attributes:
    def getattr(self):
        return ['CallName', 'ReprName', 'Pt', 'PtFunc', 'Val', 'Len',
                'LenFunc', 'Type', 'Repr', 'Trans', 'TransFunc']
    
    def showattr(self):
        for a in self.getattr():
            print('%s : %s' % (a, repr(self.__getattribute__(a))) )
    
    # cloning an Element, useful for "duplicating" an Element 
    # without keeping any dependency
    # used in Layer with Element
    # However,
    #...
    # This is not that true, as an object pointed by .Pt or .Len or .Trans
    # will not be updated to its clone()
    # Conclusion:
    # use this with care
    def clone(self):
        clone = self.__class__(
                 self.CallName, self.ReprName,
                 self.Pt, self.PtFunc,
                 self.Val, 
                 self.Len, self.LenFunc,
                 self.Repr,
                 self.Trans, self.TransFunc )
        return clone
    
    # standard method map() to map a string to the Element
    def map(self, string=''):
        if not self.is_transparent():
            l = self.map_len()
            if l is not None:
                self.Val = string[:l]
            else:
                self.Val = string
            if self.dbg >= DBG:
                log(DBG, '(Element) mapping %s on %s, %s' \
                    % (repr(string), self.CallName, repr(self)))
    
    def map_ret(self, string=''):
        self.map(string)
        return string[len(self):]
    
    # shar manipulation interface
    def to_shar(self):
        ret = shar()
        ret.set_buf(self())
        return ret
    
    def map_shar(self, sh):
        if not self.is_transparent():
            l = self.map_len()
            if l is not None:
                self.Val = sh.get_buf(l*8)
            else:
                self.Val = sh.get_buf()
            if self.dbg >= DBG:
                log(DBG, '(Element) mapping %s on %s, %s' \
                    % (repr(string), self.CallName, repr(self)))

class Int(Element):
    '''
    class defining a standard element, managed like an integer.
    It can be signed (int) or unsigned (uint),
    and support arbitrary byte-length:
        int8, int16, int24, int32, int40, int48, int56, int64, int72, ...
        uint8, uint16, uint24, uint32, ..., uint128, ...
        and why not int65536 !
    It is always byte-aligned (in term of length, at least).
    
    attributes:
    Pt: to point to another object or direct integer value;
    PtFunc: when defined, PtFunc(Pt) is used to generate the integer value;
    Val: when defined, overwrites the Pt (and PtFunc) integer value, 
         used when mapping a string buffer to the element;
    Type: type of integer for encoding, 8,16,24,32,40,48,56,64 bits signed or
          unsigned integer;
    Dict: dictionnary to use for a look-up when representing 
          the element into python;
    Repr: representation style, binary, hexa or human: human uses Dict 
          if defined;
    Trans: to define transparent element which has empty str() and len() to 0,
           it "nullifies" its existence; can point to something for automation;
    TransFunc: when defined, TransFunc(Trans) is used to automate the 
               transparency aspect: used e.g. for conditional element;
    '''
    # endianness is 'little' / 'l' or 'big' / 'b'
    _endian = 'big'
    # types format for struct library
    # 24 (16+8), 40 (32+8), 48 (32+16), 56 (32+16+8)
    _types = { 'int8':'b', 'int16':'h', 'int32':'i', 'int64':'q',
               #'int24':None, 'int40':None, 'int48':None, 'int56':None,
               'uint8':'B', 'uint16':'H', 'uint32':'I', 'uint64':'Q',
               #'uint24':None, 'uint40':None, 'uint48':None, 'uint56':None,
               }
    #
    # for object representation
    _reprs = ['hex', 'bin', 'hum']
    
    def __init__(self, CallName='', ReprName=None,
                 Pt=None, PtFunc=None, Val=None,
                 Type='int32', Dict=None, DictFunc=None,
                 Repr='hum',
                 Trans=False, TransFunc=None):
        if CallName or not self.CallName:
            self.CallName = CallName
        if ReprName is None:
            self.ReprName = ''
        else:
            self.ReprName = ReprName
        self.Pt = Pt
        self.PtFunc = PtFunc
        self.Val = Val
        self.Type = Type
        self.Dict = Dict
        self.DictFunc = DictFunc
        self.Repr = Repr
        self.Trans = Trans
        self.TransFunc = TransFunc
        # automated attributes:
        self.Len = int(self.Type.lstrip('uint'))//8
    
    def __setattr__(self, attr, val):
        # ensures no bullshit is provided into element's attributes 
        # (however, it is not a complete test...)
        if self.safe:
            if attr == 'CallName':
                if type(val) is not str or len(val) == 0:
                    raise AttributeError('CallName must be a non-null string')
            elif attr == 'ReprName':
                if type(val) is not str:
                    raise AttributeError('ReprName must be a string')
            elif attr == 'PtFunc':
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('PtFunc must be a function')
            elif attr == 'Val':
                if val is not None and not isinstance(val, (int, long)):
                    raise AttributeError('Val must be an int or long')
            elif attr == 'Type':
                cur = 0
                if val[cur] == 'u':
                    cur = 1
                if val[cur:cur+3] != 'int' or int(val[cur+3:]) % 8 != 0:
                    raise AttributeError('Type must be intX / uintX with X '\
                    'multiple of 8 bits (e.g. int24, uint32, int264, ...)')
            elif attr == 'DictFunc':
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('DictFunc must be a function')
            elif attr == 'Repr':
                if val not in self._reprs:
                    raise AttributeError('Repr %s does not exist, use in: %s' \
                          % (val, self._reprs))
            elif attr == 'TransFunc':
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('TransFunc must be a function')
        if attr == 'Type':
            self.Len = int(val.lstrip('uint'))//8
        object.__setattr__(self, attr, val)
    
    def __call__(self):
        # when no values are defined at all, arbitrary returns None:
        if self.Val is None and self.Pt is None:
            # instead of "return None" 
            # that triggers error when calling __str__() method
            return 0
        # else, Val overrides Pt capabilities:
        # transparency are not taken into account in __call__, 
        # only for the __str__ representation
        elif self.Val is not None: 
            return self.__confine( self.Val )
        elif self.PtFunc is not None: 
            if self.safe: 
                assert( type(self.PtFunc(self.Pt)) in (int, long) )
            return self.__confine(self.PtFunc(self.Pt))
        else:
            return self.__confine(int(self.Pt))
    
    def __confine(self, value):
        # unsigned
        if self.Type[0] == 'u':
            return max(0, min(2**(8*self.Len)-1, value))
        # signed
        else:
            return max(-2**(8*self.Len-1), min(2**(8*self.Len-1)-1, value))
    
    def __str__(self):
        # manages Element transparency
        if self.is_transparent():
                return ''
        # otherwise returns standard string values
        return self.__pack()
    
    def __len__(self):
        if self.is_transparent(): 
            return 0
        return self.Len
    
    def bit_len(self):
        return 8*self.__len__()
    
    # map_len() is a-priori not needed in "Int" element, 
    # but still kept for Element API uniformity
    def map_len(self):
        return self.__len__()
    
    # define integer value
    def __int__(self):
        return self()
    
    def __bin__(self):
        # unsigned or positive signed:
        val = self()
        if self.Type[0] == 'u' \
        or self.Type[0] == 'i' and val >= 0 : 
            binstr = bin(val)[2:]
            if self._endian[0] == 'l':
                # little endian
                bs = '0'*(8*self.__len__() - len(binstr)) + binstr
                bs = [bs[i:i+8] for i in range(0, len(bs), 8)]
                bs.reverse()
                return ''.join(bs)
            else:
                # big endian
                return '0'*(8*self.__len__() - len(binstr)) + binstr
        # negative signed
        else : 
            # takes 2' complement to the signed val
            binstr = bin(val + 2**(8*self.__len__()-1))[2:]
            if self._endian[0] == 'l':
                # little endian
                bs = '1' + '0'*(8*self.__len__() - len(binstr) - 1) + binstr
                bs = [bs[i:i+8] for i in range(0, len(bs), 8)]
                bs.reverse()
                return ''.join(bs)
            else:
                # big endian
                return '1' + '0'*(8*self.__len__() - len(binstr) - 1) + binstr
    
    def __hex__(self):
        return hexlify(self.__pack())
    
    def __repr__(self):
        if self.Pt is None and self.Val is None:
            return repr(None)
        if self.Repr == 'hex':
            return '0x%s' % self.__hex__()
        elif self.Repr == 'bin':
            return '0b%s' % self.__bin__()
        elif self.Repr == 'hum':
            value = self()
            if self.DictFunc:
                if self.safe:
                    assert(hasattr(self.DictFunc(self.Dict), '__getitem__'))
                try:
                    val = '%i : %s' % (value, self.DictFunc(self.Dict)[value])
                except KeyError:
                    val = value
            elif self.Dict:
                try:
                    val = '%i : %s' % (value, self.Dict[value])
                except KeyError:
                    val = value
            else:
                val = value
            rep = repr(val)
            if rep[-1] == 'L':
                return rep[:-1]
            return rep
    
    def getattr(self):
        return ['CallName', 'ReprName', 'Pt', 'PtFunc', 'Val', 'Len',
                'Type', 'Dict', 'DictFunc', 'Repr', 'Trans', 'TransFunc']
    
    def showattr(self):
        for a in self.getattr():
            if a == "Dict" and self.Dict is not None:
                print('%s : %s' % ( a, self.__getattribute__(a).__class__) )
            else:
                print('%s : %s' % ( a, repr(self.__getattribute__(a))) )
    
    def clone(self):
        clone = self.__class__(
                 self.CallName, self.ReprName,
                 self.Pt, self.PtFunc,
                 self.Val, self.Type,
                 self.Dict, self.DictFunc, self.Repr, 
                 self.Trans, self.TransFunc )
        #clone._endian = self._endian
        return clone
    
    def map(self, string=''):
        if not self.is_transparent():
            # error log will be done by the Layer().map() method
            # but do this not to throw exception
            if len(string) < self.Len:
                if self.dbg >= WNG:
                    log(WNG, '(%s) %s map(string) : string not long enough' \
                        % (self.__class__, self.CallName))
                return
            self.Val = self.__unpack(string[:self.Len])
    
    def map_ret(self, string=''):
        l = self.__len__()
        if 0 < l <= len(string):
            self.map(string)
            return string[l:]
        else:
            return string
    
    def __pack(self):
        # manage endianness (just in case...)
        if self._endian[0] == 'l':
            e = '<'
        else:
            e = '>'
        if self.Len == 0:
            return ''
        elif self.Len in (1, 2, 4, 8):
            # standard types for using Python struct
            return pack(e+self._types[self.Type], self())
        elif self.Type[0] == 'u':
            # non-standard unsigned int types
            return self.__pack_uX(e)
        else:
            # non-standard signed int types
            return self.__pack_iX(e)
    
    def __unpack(self, string):
        if self._endian[0] == 'l':
            e = '<'
        else:
            e = '>'
        if self.Len == 0:
            return
        elif self.Len in (1, 2, 4, 8):
            # standard types for using Python struct
            return unpack(e+self._types[self.Type], string[:self.Len])[0]
        elif self.Type[0] == 'u':
            # non-standard unsigned int types
            return self.__unpack_uX(string[:self.Len], e)
        else:
            # non-standard signed int types
            return self.__unpack_iX(string[:self.Len], e)
    
    def __pack_uX(self, e='>'):
        if e == '<':
            # little endian, support indefinite uint (e.g. uint3072)
            if self.Len <= 8:
                return pack('<Q', self())[:self.Len]
            else:
                #'''
                u8 = decompose(0x100, self())
                # pad with 0 until the right length
                if len(u8) < self.Len:
                    u8.extend( [0]*(self.Len-len(u8)) )
                return pack('<'+self.Len*'B', *u8)
        else:
            # big endian, support indefinite uint too
            if self.Len <= 8:
                return pack('>Q', self())[-self.Len:]
            else:
                u64 = decompose(0x10000000000000000, self())
                u64_len = 8*len(u64)
                u64_pad = 0
                if u64_len < self.Len:
                    rest_len = self.Len - u64_len
                    u64_pad = rest_len // 8
                    if rest_len % 8:
                        u64_pad += 1
                    u64.extend( [0]*u64_pad )
                u64.reverse()
                return pack('>'+len(u64)*'Q', *u64)[-self.Len:]
    
    def __pack_iX(self, e='>'):
        val = self()
        # positive values are encoded just like uint
        if val >= 0:
            return self.__pack_uX(e)
        #
        # negative values, 2's complement encoding
        if e == '<':
            # little endian
            if self.Len <= 8:
                return pack('<Q', 2**(self.Len*8) - abs(val))[:self.Len]
            else:
                i8 = decompose(0x100, 2**(self.Len*8) - abs(val))
                if len(i8) < self.Len:
                    i8.extend( [0]*(self.Len-len(i8)) )
                return pack('<'+self.Len*'B', *i8)
        else:
            # big endian
            if self.Len <= 8:
                return pack('>Q', 2**(self.Len*8) - abs(val))[-self.Len:]
            else:
                i8 = decompose(0x100, 2**(self.Len*8) - abs(val))
                if len(i8) < self.Len:
                    i8.extend( [0]*(self.Len-len(i8)) )
                i8.reverse()
                return pack('>'+self.Len*'B', *i8)
    
    def __unpack_uX(self, string, e='>'):
        if e == '<':
            # little endian, support indefinite uint (e.g. uint3072)
            if self.Len <= 8:
                return unpack('<Q', string + '\0'*(8 - self.Len))[0]
            else:
                u8 = list(unpack('<'+self.Len*'B', string))
                u8.reverse()
                return reduce(lambda x,y: (x<<8)+y, u8)
        else:
            # big endian, support indefinite uint (e.g. uint3072)
            if self.Len <= 8:
                return unpack('>Q', '\0'*(8 - self.Len) + string)[0]
            else:
                #'''
                u64_len, u64_pad = self.Len // 8, self.Len % 8
                if u64_pad:
                    u64_len += 1
                    string = '\0'*u64_pad + string
                u64 = unpack('>'+u64_len*'Q', string)
                return reduce(lambda x,y:(x<<64)+y, u64)
    
    def __unpack_iX(self, string, e='>'):
        if e == '<':
            # little endian
            if ord(string[-1]) & 0x80 == 0:
                # check if it is a positive value
                return self.__unpack_uX(string, e)
            elif self.Len <= 8:
                return unpack('<Q', string + '\0'*(8 - self.Len))[0] \
                       - 2**(self.Len*8)
            else:
                i8 = list(unpack('<'+self.Len*'B', string))
                i8.reverse()
                return reduce(lambda x,y:(x<<8)+y, i8) - 2**(self.Len*8)
        else:
            # big endian
            if ord(string[0]) & 0x80 == 0:
                # check if it is a positive value
                return self.__unpack_uX(string, e)
            elif self.Len <= 8:
                return unpack('>Q', '\0'*(8 - self.Len) + string)[0] \
                       - 2**(self.Len*8)
            else:
                i8 = unpack('>'+self.Len*'B', string)
                return reduce(lambda x,y:(x<<8)+y, i8) - 2**(self.Len*8)
    
    # shar manipulation interface
    def to_shar(self):
        ret = shar()
        ret.set_buf(self.__str__())
        return ret
    
    def map_shar(self, sh):
        if len(sh)*8 < self.Len:
            if self.dbg >= WNG:
                log(WNG, '(%s) %s map(string) : shar buffer not long enough' \
                    % (self.__class__, self.CallName))
            return
        # standard handling
        if not self.is_transparent():
            self.map(sh.get_buf(len(self)*8))
    
class Bit(Element):
    '''
    class defining a standard element, managed like a bit (e.g. a flag)
    or bit-stream of variable bit length
    Values are corresponding to unsigned integer: from 0 to pow(2, bit_len)-1.
    It does not require to be byte-aligned.
    
    attributes:
    Pt: to point to another object or direct integer value;
    PtFunc: when defined, PtFunc(Pt) is used to generate the integer value;
    Val: when defined, overwrites the Pt (and PtFunc) integer value, 
         used when mapping string to the element;
    BitLen: length in bits of the bit stream;
    BitLenFunc: to be used when mapping string with variable bit-length, 
                BitLenFunc(BitLen) is used;
    Dict: dictionnary to use for a look-up when representing 
          the element into python;
    Repr: representation style, binary, hexa or human: human uses Dict;
    Trans: to define transparent element which has empty str() and len() to 0,
           it "nullifies" its existence; can point to something for automation;
    TransFunc: when defined, TransFunc(Trans) is used to automate the 
               transparency aspect: used e.g. for conditional element;
    '''
    # for object representation
    _reprs = ['hex', 'bin', 'hum']
    
    def __init__(self, CallName='', ReprName=None, 
                 Pt=None, PtFunc=None, Val=None, 
                 BitLen=1, BitLenFunc=None,
                 Dict=None, DictFunc=None, Repr='bin', 
                 Trans=False, TransFunc=None):
        if CallName or not self.CallName:
            self.CallName = CallName
        if ReprName is None: 
            self.ReprName = ''
        else: 
            self.ReprName = ReprName
        self.Pt = Pt
        self.PtFunc = PtFunc
        self.Val = Val
        self.BitLen = BitLen
        self.BitLenFunc = BitLenFunc
        self.Dict = Dict
        self.DictFunc = DictFunc
        self.Repr = Repr
        self.Trans = Trans
        self.TransFunc = TransFunc
    
    def __setattr__(self, attr, val):
        # ensures no bullshit is provided into element's attributes 
        # (however, it is not a complete test...)
        if self.safe:
            if attr == 'CallName':
                if type(val) is not str or len(val) == 0:
                    raise AttributeError('CallName must be a non-null string')
            elif attr == 'ReprName':
                if type(val) is not str:
                    raise AttributeError('ReprName must be a string')
            elif attr == 'PtFunc':
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('PtFunc must be a function')
            elif attr == 'Val':
                if val is not None and not isinstance(val, (int, long)):
                    raise AttributeError('Val must be an int')
            elif attr == 'BitLenFunc':
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('BitLenFunc must be a function')
            elif attr == 'DictFunc':
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('DictFunc must be a function')
            elif attr == 'Repr':
                if val not in self._reprs:
                    raise AttributeError('Repr %s does not exist, use in: %s' \
                          % (val, self._reprs))
            elif attr == 'TransFunc':
                if val is not None and not isinstance(val, type_funcs) :
                    raise AttributeError('TransFunc must be a function')
        object.__setattr__(self, attr, val)
    
    def __call__(self):
        if self.Val is None and self.Pt is None: return 0
        if self.Val is not None: return self.__confine(self.Val) 
        elif self.PtFunc is not None:
            if self.safe:
                assert( type(self.PtFunc(self.Pt)) is int )
            return self.__confine(self.PtFunc(self.Pt))
        else: return self.__confine(int(self.Pt))
    
    def __confine(self, value):
        # makes sure value provided does not overflow bit length
        return max( 0, min( pow(2, self.bit_len())-1, value ))
    
    def __str__(self):
        # return the string representation of the integer value
        # big-endian encoded
        # left-aligned according to the bit length
        # -> last bits of the last byte are nullified
        # 
        # manages Element transparency
        if self.is_transparent():
            return ''
        # do it the dirty way:
        h = self.__hex__()
        if not h:
            return ''
        if len(h) % 2: h = ''.join(('0', h))
        return str(shtr(unhexlify(h)) << (8-(self.bit_len()%8))%8)
    
    def _shar__str__(self):
        bitlen = self.bit_len()
        if bitlen == 0:
            return ''
        ret = shar()
        ret.set_uint(self(), bitlen)
        return ret.to_buf()
    
    def __len__(self):
        # just for fun here, 
        # but do not use this in program...
        bitlen = self.bit_len()
        if bitlen % 8:
            return (bitlen//8) + 1
        return bitlen//8
    
    def bit_len(self):
        # manages Element transparency
        if self.is_transparent():
            return 0
        # and standard bit length processing
        if self.BitLenFunc is not None:
            if self.safe:
                assert( type(self.BitLenFunc(self.BitLen)) is int )
            return self.BitLenFunc(self.BitLen)
        else:
            if self.safe:
                assert( type(self.BitLen) is int )
            return self.BitLen
    
    def map_len(self):
        bitlen = self.bit_len()
        if bitlen % 8:
            return (bitlen//8)+1
        return bitlen//8
    
    def __hex__(self):
        bitlen = self.bit_len()
        if not bitlen:
            return ''
        hexa = hex(self())[2:]
        if hexa[-1] == 'L':
            # thx to Python to add 'L' for long on hex repr...
            hexa = hexa[:-1]
        if self.bit_len()%4: 
            return '0'*(self.bit_len()//4 + 1 - len(hexa)) + hexa
        else: 
            return '0'*(self.bit_len()//4 - len(hexa)) + hexa
    
    def __int__(self):
        return self()
        
    def __bin__(self):
        bitlen = self.bit_len()
        if not bitlen:
            return ''
        binary = format(self(), 'b')
        return (bitlen - len(binary))*'0' + binary
        
    def __repr__(self):
        if self.Repr == 'hex': return '0x%s' % self.__hex__()
        elif self.Repr == 'bin': return '0b%s' % self.__bin__()
        elif self.Repr == 'hum':
            value = self()
            if self.DictFunc:
                if self.safe:
                    assert(hasattr(self.DictFunc(self.Dict), '__getitem__'))
                try: val = '%i : %s' % (value, self.DictFunc(self.Dict)[value])
                except KeyError: val = value
            elif self.Dict:
                try: val = '%i : %s' % (value, self.Dict[value])
                except KeyError: val = value
            else:
                val = value
            #return repr(val)
            rep = repr(val)
            if rep[-1] == 'L':
                rep = rep[:-1]
            return rep
    
    def getattr(self):
        return ['CallName', 'ReprName', 'Pt', 'PtFunc', 'Val', 'BitLen',
                'BitLenFunc', 'Dict', 'DictFunc', 'Repr', 'Trans', 'TransFunc']
    
    def showattr(self):
        for a in self.getattr():
            if a == 'Dict' and self.Dict is not None: 
                print('%s : %s' % ( a, self.__getattribute__(a).__class__) )
            else: 
                print('%s : %s' % ( a, repr(self.__getattribute__(a))) )
    
    # cloning an element, used in set of elements
    def clone(self):
        clone = self.__class__(
                 self.CallName, self.ReprName,
                 self.Pt, self.PtFunc,
                 self.Val, 
                 self.BitLen, self.BitLenFunc,
                 self.Dict, self.DictFunc, self.Repr,
                 self.Trans, self.TransFunc )
        return clone
    
    def map(self, string=''):
        # map each bit of the string from left to right
        # using the shtr() class to shift the string
        # string must be ascii-encoded (see shtr)
        if not self.is_transparent():
            self.map_bit( shtr(string).left_val(self.bit_len()) )
    
    def map_bit(self, value=0):
        # map an int / long value
        if self.safe:
            assert( 0 <= value <= pow(2, self.bit_len()) )
        self.Val = value
    
    def map_ret(self, string=''):
        if self.is_transparent():
            return string
        else:
            shtring = shtr(string)
            bitlen = self.bit_len()
            self.map_bit( shtring.left_val(bitlen) )
            return shtring << bitlen
    
    # shar manipulation interface
    def to_shar(self):
        ret = shar()
        ret.set_uint(self(), self.bit_len())
        return ret
    
    def map_shar(self, sh):
        if not self.is_transparent():
            self.map_bit( sh.get_uint(self.bit_len()) )
    
#------------------------------------------------------------------------------#
# Layer definition
#------------------------------------------------------------------------------#
class Layer(object):
    '''
    class built from stack of "Str", "Int", "Bit" and "Layer" objects
    got from the initial constructorList.
    Layer object is recursive: it can contain other Layer() instances
    Layer does not require to be byte-aligned. This happens depending of the
    presence of Bit() instances.
    
    when instantiated:
    clones the list of "Str", "Int", "Bit" elements in the constructorList
    to build a dynamic elementList, that can be changed afterwards (adding /
    removing objects);
    A common hierarchy level for the whole Layer is defined, it is useful 
    when used into "Block" to create hierarchical relationships: 
        self.hierarchy (int), self.inBlock (bool)
        when .inBlock is True, provides: .get_payload(), .get_header(), 
             .has_next(), .get_next(), .get_previous(), and .Block
    It provides several methods for calling elements in the layer:
        by CallName / ReprName passed in attribute
        by index in the elementList
        can be iterated too
        and many other manipulations are defined
    It has also some common methods with "Str", "Int" and "Bit" to emulate 
    a common handling:
    __str__, __len__, __int__, bit_len, getattr, showattr, show, map
    '''
    #
    # debugging threshold for Layer:
    dbg = ERR
    # add some sanity checks
    #safe = True
    safe = False
    # define the type of str() and map() method
    _byte_aligned = True
    # reserved attributes:
    Reservd = ['CallName', 'ReprName', 'elementList', 'Len', 'BitLen',
               'hierarchy', 'inBlock', 'Trans', 'ConstructorList',
               'dbg', 'Reservd']
    #
    # represent transparent elements in __repr__()
    _repr_trans = True
    
    # structure description:
    constructorList = []
    
    def __init__(self, CallName='', ReprName='', Trans=False, **kwargs):
        if type(CallName) is not str:
            raise AttributeError('CallName must be a string')
        elif len(CallName) == 0:
            self.CallName = self.__class__.__name__
        else:
            self.CallName = CallName
        if type(ReprName) is str and len(ReprName) > 0: 
            self.ReprName = ReprName
        else: 
            self.ReprName = ''
        self.elementList = []
        self.set_hierarchy(0)
        self.inBlock = False
        self.Trans = Trans
        
        CallNames = []
        for e in self.constructorList:
            # This is for little players
            #if isinstance(e, Element):
            # OK, now let's put the balls on the table and
            # make Layer recursive (so will have Layer() into Layer())
            if isinstance(e, (Element, Layer)):
                if e.CallName in self.Reservd:
                    if self.safe or self.dbg >= ERR:
                        log(ERR, '(Layer - %s) using a reserved '
                            'attribute as CallName %s: aborting...' \
                          % (self.__class__, e.CallName))
                    return
                if e.CallName in CallNames:
                    if self.dbg >= WNG:
                        log(WNG, '(Layer - %s) different elements have ' \
                           'the same CallName %s' % (self.__class__, e.CallName))
                if isinstance(e, Element):
                    self.append(e.clone())
                # do not clone Layer() as it breaks dynamic element inside
                # i.e. element with PtFunc, LenFunc, DictFunc, TransFunc defined
                # TODO: patch Layer().clone() method to solve this...
                # lets try with deepcopy()
                elif isinstance(e, Layer):
                    self.append(e.clone())
            CallNames.append(e.CallName)
        
        # check for bit alignment until we lost information on the Layer length
        # also check if fixed length can be deduced
        self.BitLen = 0
        for e in self.elementList:
            if self.dbg >= DBG:
                log(DBG, '(Layer - %s) length verification for %s' \
                    % (self.__class__, e.CallName))
            if isinstance(e, Bit):
                self.BitLen += e.bit_len()
            elif hasattr(e, 'Len') and type(e.Len) is int:
                self.BitLen += (e.Len)*8
            else:
                self.BitLen, self.Len = 'var', 'var'
                break
        if type(self.BitLen) is int :
            if self.BitLen % 8:
                if self.dbg >= WNG and self._byte_aligned:
                    log(WNG, '(Layer - %s) Elements seem not to be '\
                        'byte-aligned: hope you expect it!' \
                        % self.__class__)
                # record length in bit (precise one) and in bytes (unprecised)
                self.Len = 1 + self.BitLen//8
            else:
                self.Len = self.BitLen//8
        #
        # check additional args that would correspond to contained Element
        args = kwargs.keys()
        if self.dbg >= DBG:
            log(DBG, '(Layer - %s) init kwargs: %s' % (self.__class__, args))
        for e in self:
            if hasattr(e, 'CallName') and hasattr(e, 'Pt') \
            and e.CallName in args:
                e.Pt = kwargs[e.CallName]
    
    # define some basic list facilities for managing elements into the Layer, 
    # through the "elementList" attribute:
    def __iter__(self):
        if 'elementList' in self.__dict__.keys():
            return self.__dict__['elementList'].__iter__()
        else: return [].__iter__()
    
    def __getitem__(self, num):
        return self.elementList[num]
    
    def __getslice__(self, i, j):
        l = Layer('_slice_')
        if not i or i < 0:
            i=0
        #maxj = len(self.elementList)-1
        maxj = len(self.elementList)
        if not j or j > maxj:
            j = maxj
        #
        for k in xrange(i, j):
            l.append( self[k] )
        return l
    
    def __setitem__(self, num, value):
        # special handling here: 
        # use to override the element value 
        # with its "Val" attribute (like when mapping a string)
        self.elementList[num].Val = value
    
    def append(self, element):
        #if isinstance(element, Element):
        # make Layer recursive:
        if isinstance(element, (Element, Layer)):
            if self.dbg >= WNG and element.CallName in self.getattr():
                log(WNG, '(Layer - %s) different elements have the same '\
                         'CallName %s' % (self.__class__, element.CallName))
            self.elementList.append(element)
    
    def __lshift__(self, element):
        self.append(element)
        if isinstance(element, Layer):
            element.inc_hierarchy(self.hierarchy)
    
    def insert(self, index, element):
        CallNames = self.getattr()
        #if isinstance(element, Element):
        # make Layer recursive:
        if isinstance(element, (Element, Layer)):
            if self.dbg >= WNG and element.CallName in CallNames:
                log(WNG, '(Layer - %s) different elements have the same '\
                         'CallName %s' % (self.__class__, element.CallName))
            self.elementList.insert(index, element)
    
    def __rshift__(self, element):
        self.insert(0, element)
        if isinstance(element, Layer):
            element.inc_hierarchy(self.hierarchy)
    
    def extend(self, newElementList):
        for e in newElementList:
            self.append(e)
    
    def remove(self, element):
        for e in self:
            if e == element:
                self.elementList.remove(element)
    
    def replace(self, current_element, new_element):
        # check index of the element ro replace
        index = 0
        for elt in self.elementList:
            if elt == current_element:
                self.remove(current_element)
                self.insert(index, new_element)
                return
            else:
                index += 1
    
    # define some attribute facilities for managing elements 
    # by their CallName into the Layer
    # warning: dangerous when parsing data into Layer, 
    # with elements which could have same CallName
    # 
    # list facilities can be preferred in this case
    def __getattr__(self, name):
        names = [x.CallName for x in self.elementList]
        if name in names:
            return self.elementList[ names.index(name) ]
        names = [x.ReprName for x in self.elementList]
        if name in names:
            return self.elementList[ names.index(name) ]
        #
        return object.__getattribute__(self, name)
        #return self.__getattribute__(name)
        #return getattr(self, name)
    
    def __setattr__(self, name, value):
        # special handling here: use to override the element value 
        # with its "Val" attribute (like when mapping a string)
        for e in self:
            if name == e.CallName or name == e.ReprName: 
                e.Val = value
                return
        return object.__setattr__(self, name, value)
        raise AttributeError( '"Layer" has no "%s" attribute: %s' \
              % (name, self.getattr()) )
    
    def __hasattr__(self, name):
        for e in self:
            if name == e.CallName or name == e.ReprName: 
                return True
        #return object.__hasattr__(self, name): 
        # not needed (does not work in the code... but works in python...)
        raise AttributeError( '"Layer" has no "%s" attribute: %s' \
              % (name, self.getattr()) )
    
    # method for managing the Layer hierarchy (easy):
    def set_hierarchy(self, hier=0):
        self.hierarchy = hier
        for e in self:
            if isinstance(e, Layer):
                e.set_hierarchy(hier)
        
    def inc_hierarchy(self, ref=None):
        if ref is None:
            self.set_hierarchy(self.hierarchy+1)
        else: 
            self.set_hierarchy(self.hierarchy+ref+1)
        #for l in self:
        #    if isinstance(l, Layer):
        #        l.hierarchy = self.hierarchy
    
    def dec_hierarchy(self, ref=None):
        if ref is None: 
            self.set_hierarchy(self.hierarchy-1)
        else: 
            self.set_hierarchy(self.hierarchy+ref-1)
        #for l in self:
        #    if isinstance(l, Layer):
        #        l.hierarchy = self.hierarchy
    
    # define same methods as "Element" type for being use the same way
    def __str__(self):
        if self.dbg >= DBG:
            log(DBG, '(Layer.__str__) entering str() for %s' % self.CallName)
        # First take care of transparent Layer (e.g. in L3Mobile)
        if hasattr(self, 'Trans') and self.Trans:
            return ''
        # dispatch to the right method depending of byte alignment
        if self._byte_aligned is True:
            return self.__str_aligned()
        else:
            return self.__str_unaligned()
    
    def __str_unaligned(self):
        # then init resulting string 
        # and bit offset needed to shift unaligned strings
        s = []
        off = 0
        # loop on each element into the Layer
        # also on Layer into Layer...
        for e in self:
            shtr_e, bitlen_e = e.shtr(), e.bit_len()
            rest_e = bitlen_e % 8
            if self.dbg >= DBG:
                log(DBG, '(Layer.__str__) %s: %s, %i, offset: %i' \
                    % (e.CallName, hexlify(shtr_e), bitlen_e, off))
            # if s is not byte-aligned and e is not transparent
            # update the last component of s
            if off and shtr_e:
                # 1st update last bits of s with MSB of e, 
                # before stacking with the rest of e
                # (8 - off) is the room left in the LSB of s
                s[-1] = ''.join((s[-1][:-1],
                                  chr(ord(s[-1][-1]) + shtr_e.left_val(8-off)),
                                  (shtr_e << (8-off))
                                ))
                # take care in case the shifting of e voids its last byte
                if rest_e and rest_e-(8-off) <= 0:
                    s[-1] = s[-1][:-1]
                # update offset
                if rest_e:
                    off += bitlen_e
                    off %= 8
            # in case s is already aligned, append a new component to s
            elif shtr_e:
                s.append(shtr_e)
                # update offset
                if rest_e:
                    off += rest_e
            if self.dbg >= DBG:
                log(DBG, '(Layer.__str__) %s' % hexlify(s))
        # well done!
        return ''.join(s)
    
    def __str_aligned(self):
        s = []
        BitStream = ''
        # loop on each element in the Layer
        # also on Layer into Layer...
        for e in self:
            # need special processing for stacking "Bit" element: 
            #   using "BitStream" variable
            #   works only with contiguous "Bit" elements 
            #   to avoid byte-misalignment of other element in the Layer 
            #   (and programming complexity with shifting everywhere...)
            if isinstance(e, Bit):
                # manage element transparency with (Trans, TranFunc)
                # and build a bitstream ('100110011...1101011') from Bit values
                if not e.is_transparent():
                    BitStream += str(e.__bin__())
                # when arriving on a byte boundary from bitstream, 
                # create bytes and put it into the s variable
                if len(BitStream) >= 8:
                    while 1:
                        s.append( pack('!B', int(BitStream[:8], 2)) )
                        BitStream = BitStream[8:]
                        if len(BitStream) < 8:
                            break
                if self.dbg >= DBG:
                    log(DBG, '(Element) %s: %s, %s\nBitstream: %s' \
                        % (e.CallName, e(), e.__bin__(), BitStream))
            # when going to standard Str or Int element, 
            # or directly end of __str__ function 
            # verify the full BitStream has been consumed
            # and continue to build the resulting string easily...
            else:
                # possible byte mis-alignment for Str / Int is not managed...
                self.__is_aligned(BitStream)
                BitStream = ''
                if isinstance(e, Layer) and not e.Trans \
                or isinstance(e, Element):
                    s.append( str(e) )
        self.__is_aligned(BitStream)
        return ''.join(s)
    
    def __is_aligned(self, BitStream):
        if BitStream and self.dbg >= ERR:
            log(ERR, '(Layer - %s) some of the Bit elements have not been ' \
                'stacked in the "str(Layer)"\nremaining bitstream: %s' \
                % (self.__class__, BitStream))
            if self.safe:
                assert(not BitStream)
    
    def __call__(self):
        return self.__str__()
    
    def __len__(self):
        return len(self.__str__())
    
    def shtr(self):
        return shtr(self.__str__())
    
    def bit_len(self):
        # just go over all internal elements to track their own bit length
        # updated attributes initialized when Layer was constructed
        self.BitLen = 0
        for e in self:
            if hasattr(e, 'bit_len'):
                self.BitLen += e.bit_len()
            elif hasattr(e, '__len__'):
                self.BitLen += len(e)*8
        self.Len = 1 + (self.BitLen // 8) if self.BitLen % 8 \
                   else (self.BitLen // 8)
        return self.BitLen
    
    def __hex__(self):
        bit_len = self.bit_len()
        hex_len = bit_len/4
        if bit_len%4:
            hex_len += 1
        #
        return self.__str__().encode('hex')[:hex_len]
    
    def __bin__(self):
        bits = []
        for e in self:
            bits.append( e.__bin__() )
        return ''.join(bits)
    
    def __int__(self):
        # big endian integer representation of the string buffer
        if self._byte_aligned:
            return shtr(self).left_val(len(self)*8)
        else:
            return shtr(self).left_val(self.bit_len())
    
    def __repr__(self):
        t = ''
        if self.Trans:
            t = ' - transparent '
        s = '<%s[%s]%s: ' % ( self.ReprName, self.CallName, t )
        for e in self:
            if self._repr_trans or not e.is_transparent():
                s += '%s(%s):%s, ' % ( e.CallName, e.ReprName, repr(e) )
        s = s[:-2] + '>'
        return s
    
    def map_len(self):
        return len(self)
    
    def getattr(self):
        return [e.CallName for e in self.elementList]
    
    def showattr(self):
        for a in self.getattr():
            print('%s : %s' % ( a, repr(self.__getattr__(a))) )
    
    def clone(self):
        return deepcopy(self)
    
    def is_transparent(self):
        if self.Trans:
            return True
        else:
            return False
    
    def show(self, with_trans=False):
        re, tr = '', ''
        if self.ReprName != '':
            re = '%s ' % self.ReprName
        if self.is_transparent():
            # TODO: eval the best convinience here
            if not with_trans:
                return ''
            tr = ' - transparent'
        # Layer content
        str_lst = [e.show().replace('\n', '\n ') for e in self]
        # insert spaces for nested layers and filter out empty content
        str_lst = [' %s\n' % s for s in str_lst if s]
        # insert layer's title
        str_lst.insert(0, '### %s[%s]%s ###\n' % (re, self.CallName, tr))
        # return full inline string without last CR
        return ''.join(str_lst)[:-1]
    
    def map(self, string=''):
        if self.dbg >= DBG:
            log(DBG, '(Layer.map) entering map() for %s' % self.CallName)
        # First take care of transparent Layer (e.g. in L3Mobile)
        if hasattr(self, 'Trans') and self.Trans:
            return
        # dispatch to the right method depending of byte alignment
        if self._byte_aligned is True:
            self.__map_aligned(string)
        else:
            self.__map_unaligned(string)
    
    def __map_unaligned(self, string=''):
        s = shtr(string)
        # otherwise go to map() over all elements
        for e in self:
            if self.dbg >= DBG:
                log(DBG, '(Layer.__map_unaligned) %s, bit length: %i' \
                    % (e.CallName, e.bit_len()))
                log(DBG, '(Layer.__map_unaligned) string: %s' % hexlify(s))
            # this is beautiful
            s = e.map_ret(s)
    
    def __map_aligned(self, string=''):
        # Bit() elements are processed intermediary: 
        # 1st placed into BitStack
        # and when BitStack is byte-aligned (check against BitStack_len)
        # string buffer is then mapped to it
        self.__BitStack = []
        self.__BitStack_len = 0
        # Furthermore, it manages only contiguous Bit elements 
        # for commodity... otherwise, all other elements should be shifted
        #
        for e in self:
            # special processing for Bit() element:
            if isinstance(e, Bit):
                self.__add_to_bitstack(e)
                # if BitStack is byte aligned, map string to it:
                if self.__BitStack_len % 8 == 0:
                    string = self.__map_to_bitstack(string)
            # for other elements (Str(), Int(), Layer()), standard processing:   
            else:
                if self.__BitStack_len > 0 and self.dbg >= ERR:
                    log(WNG, '(Layer - %s) some of the Bit elements have not ' \
                        'been mapped in the "Layer": not byte-aligned' \
                        % self.__class__)
                if isinstance(e, (Layer, Element)) and not e.is_transparent():
                    if len(string) < e.map_len() and self.dbg >= WNG:
                        log(WNG, '(Layer - %s) String buffer not long ' \
                            'enough for %s' % (self.__class__, e.CallName))
                        #if self.safe:
                        #    return
                    e.map(string)
                    string = string[e.map_len():]
        # delete .map() *internal* attributes
        del self.__BitStack
        del self.__BitStack_len
    
    def __add_to_bitstack(self, bit_elt):
        # check for Bit() element transparency
        if not bit_elt.is_transparent():
            self.__BitStack += [bit_elt]
            self.__BitStack_len += bit_elt.bit_len()
            
    def __map_to_bitstack(self, string):
        # 1st check if string is long enough for the prepared BitStack
        if len(string) < self.__BitStack_len//8 and self.dbg >= ERR:
            log(ERR, '(Layer - %s) String buffer not long enough for %s' \
                % (self.__class__, self.__BitStack[-1].CallName))
            #if self.safe:
            #    return
        # string buffer parsing is done through intermediary
        # string buffer "s_stack"
        s_stack = string[:self.__BitStack_len//8]
        # create a bitstream "s_bin" for getting the full BitStack
        s_bin = ''
        for char in s_stack:
            # convert to bitstream thanks to python native bit repr
            s_bin_tmp = bin(ord(char))[2:]
            # prepend 0 to align on byte (python does not do it)
            # and append to the bitstream "s_bin" (string of 0 and 1)
            s_bin = ''.join((s_bin, (8-len(s_bin_tmp))*'0', s_bin_tmp))
        # map the bitstream "s_bin" into each BitStack element
        for bit_elt in self.__BitStack:
            bitlen = bit_elt.bit_len()
            if bitlen:
                # convert the bitstream "s_bin" into integer 
                # according to the length in bit of bit_elt
                bit_elt.map_bit( int(s_bin[:bit_elt.bit_len()], 2) )
                # truncate the "s_bin" bitstream
                s_bin = s_bin[bit_elt.bit_len():]
        # consume the global string buffer that has been mapped 
        # (from s_stack internal variable)
        # and reinitialize self.__BitStack* attributes
        string = string[self.__BitStack_len//8:]
        self.__BitStack = []
        self.__BitStack_len = 0
        # finally return string to parent method .map()
        return string
    
    # map_ret() maps a buffer to a Layer, the unaligned way,
    # and returns the rest of the buffer that was not mapped
    def map_ret(self, string=''):
        if self.dbg >= DBG:
            log(DBG, '(Layer.map_ret) entering map_ret() for %s' % self.CallName)
        # First take care of transparent Layer (e.g. in L3Mobile)
        if hasattr(self, 'Trans') and self.Trans:
            return string
        if self._byte_aligned is True:
            self.__map_aligned(string)
            return string[len(self):]
        else:
            # actually, map_ret() is only interesting for unaligned layers
            s = shtr(string)
            for e in self:
                if self.dbg >= DBG:
                    log(DBG, '(Layer.map_ret) %s, bit length: %i' \
                        % (e.CallName, e.bit_len()))
                    log(DBG, '(Layer.map_ret) string: %s' % hexlify(s))
                # this is beautiful
                s = e.map_ret(s)
            return s
    
    # define methods when Layer is in a Block:
    # next, previous, header: return Layer object reference
    # payload: returns Block object reference
    def get_index(self):
        if self.inBlock is not True: 
            return 0
        i = 0
        for l in self.Block:
            if l == self: return i
            else:  i += 1
        # could happen if Layer is placed in several Block()
        # and the last Block used is deleted
        return False
    
    def has_next(self):
        if self.inBlock is not True: 
            return False
        return self.Block.has_index(self.get_index()+1)
    
    def get_next(self):
        if self.has_next():
            return self.Block[self.get_index()+1]
        return RawLayer()
    
    def has_previous(self):
        if self.inBlock is not True: 
            return False
        index = self.get_index()
        if index <= 0:
            return False
        return self.Block.has_index(index-1)
        
    def get_previous(self):
        if self.has_previous():
            return self.Block[self.get_index()-1]
        return RawLayer()
        
    def get_header(self):
        if self.has_previous():
            index = self.get_index()
            i = index - 1
            while i >= 0:
                if self.Block[i].hierarchy == self.hierarchy-1:
                    return self.Block[i]
                else:
                    i -= 1
        return RawLayer()
    
    def get_payload(self):
        # return a Block, not a Layer like other methods 
        # for management into a Block
        pay = Block('pay')
        if self.has_next():
            index = self.get_index()
            for l in self.Block[ index+1 : ]:
                if l.hierarchy > self.hierarchy:
                    #pay.append( l.clone() )
                    # not needed to append a clone
                    # better keep reference to original layer
                    pay.append( l )
                else:
                    break
            if pay.num() == 0:
                pay.append( RawLayer() )
            return pay
        pay.append( RawLayer() )
        return pay
    
    def num(self):
        return 1
    
    # this is to retrieve full Layer's dynamicity from a mapped layer
    def reautomatize(self):
        for e in self:
            if hasattr(e, 'reautomatize'):
                e.reautomatize()
    
    def parse(self, s=''):
        self.map(s)
    
    # shar manipulation interface
    def to_shar(self):
        if hasattr(self, 'Trans') and self.Trans:
            return shar()
        bits = []
        for e in self:
            bits.append( e.to_shar().get_bits() )
        s = shar()
        s.set_bits( bits )
        return s
    
    def map_shar(self, sh):
        if self.dbg >= DBG:
            log(DBG, '(Layer.map) entering map() for %s' % self.CallName)
        # First take care of transparent Layer (e.g. in L3Mobile)
        if hasattr(self, 'Trans') and self.Trans:
            return
        if not isinstance(sh, shar):
            sh = shar(sh)
        for e in self:
            e.map_shar(sh)


class RawLayer(Layer):
    constructorList = [
        Str(CallName='s', Pt='', Len=None),
        ]
    
    def __init__(self, s=''):
        Layer.__init__(self, CallName='raw')
        self.s.Pt = s

#------------------------------------------------------------------------------#
# Block definition
#------------------------------------------------------------------------------#
class Block(object):
    '''
    class to build a block composed of "Layer" objects
    define only methods, not content, which depends of the protocol or data model.
    
    when instantiated:
    defines a layerList;
    has methods to append, remove, extend, insert... layers into the block;
    provides several methods for calling layers in the block;
    to manage their hierarchy...;
    and use of signs such as >> << | to include layers into the block;
    
    have also same methods as "Layer": map, show, clone
    to emulate a common handling.
    '''
    # debugging thresholf for Block:
    dbg = 0
    
    def __init__(self, Name=''):
        if type(Name) is not str:
            raise AttributeError('CallName must be a string')
        self.CallName = Name
        self.layerList = []
        self.set_hierarchy(0)
        self.inBlock = False

    # define some basic list facilities for managing layers into the Block:
    def __iter__(self):
        if 'layerList' in self.__dict__.keys():
            return self.__dict__['layerList'].__iter__()
        else: return [].__iter__()
    
    def __getitem__(self, num):
        return self.layerList[num]
    
    def __getattr__(self, attr):
        for l in self:
            if attr == l.CallName: return l
        return object.__getattribute__(self, attr)
    
    def num(self):
        return len(self.layerList)
    
    def has_index(self, index):
        if index < self.num(): return True
        return False
    
    def append(self, obj):
        #if isinstance(obj, Layer):
        # and the following will be great !!!
        # this lib starts really to amaze me!
        if isinstance(obj, (Layer, Block)):
            obj.inBlock = True
            # do not re-assign Layer.Block when building a payload
            # referring to Layer.get_payload()
            if self.CallName != 'pay':
                obj.Block = self
            # better keep original Layer / Block hierarchy
            #obj.hierarchy = self.hierarchy
            self.layerList.append(obj)
    
    def extend(self, block):
    # would need some more intelligence...
        if isinstance(block, Block):
            self.layerList.extend(block.layerList)
    
    def insert(self, index, layer):
        if isinstance(layer, Layer):
            layer.inBlock = True
            layer.Block = self
            self.layerList.insert(index, layer)
    
    def remove(self, start, stop=None):
        if stop is None:
            self.layerList.remove( self.layerList[start] )
        else:
            for i in xrange(start, stop):
                self.layerList.remove( self.layerList[start] )
    
    # method for Block hierarchy setting
    def set_hierarchy(self, hier):
        self.hierarchy = hier
        for l in self:
            l.set_hierarchy(l.hierarchy + hier)
    
    # method for managing all the Layers hierarchy in the Block (easy):
    def inc_hierarchy(self, ref=0):
        #self.hierarchy += 1+ref
        #for l in self: l.hierarchy += 1+ref
        self.set_hierarchy(self.hierarchy+1+ref)
        
    def dec_hierarchy(self, ref=0):
        #self.hierarchy -= 1-ref
        #for l in self: l.hierarchy -= 1-ref
        self.set_hierarchy(self.hierarchy-1+ref)
    
    # define operations to insert layers into a block:
    # OR: block | new_layer, append the new_layer with the same hierarchy 
    # as last layer in the block
    def __or__(self, newLayer):
        self.append(newLayer)
        self[-1].set_hierarchy(self[-2].hierarchy)
        #for l in self[-1]:
        #    if isinstance(l, Layer):
        #        l.hierarchy = self[-1].hierarchy
    
    # LSHIFT: block << new_layer, append the new_layer with a higher hierarchy
    # than last layer in the block
    def __lshift__(self, newLayer):
        self.append(newLayer)
        if self.num() > 1:
            self[-1].inc_hierarchy( self[-2].hierarchy )
        else:
            self[-1].set_hierarchy(self.hierarchy)
    
    # RSHIFT: block >> new_layer, append the new_layer with a lower hierarchy
    # than last layer in the block
    def __rshift__(self, newLayer):
        self.append(newLayer)
        self[-1].dec_hierarchy( self[-2].hierarchy )
    
    # standard methods for common management with Layers
    def __str__(self):
        s = []
        for l in self:
            if not hasattr(self, 'Trans') or not l.Trans:
                s.append( str(l) )
        return ''.join(s)
    
    def shtr(self):
        return shtr(self.__str__())
    
    def __len__(self):
        return len(self.__str__())
    
    def __int__(self):
        # big endian integer representation of the string buffer
        if self._byte_aligned:
            return shtr(self).left_val(len(self)*8)
        else:
            return shtr(self).left_val(self.bit_len())
    
    def __repr__(self):
        s = '[[%s] ' % self.CallName
        for l in self:
            s += l.__repr__()
        s = s + ' [%s]]' % self.CallName
        return s
    
    def map_len(self):
        return len(self)
    
    def clone(self):
        clone = self.__class__()
        clone.CallName, clone.layerList = self.CallName, []
        for l in self:
            if isinstance(l, Layer): 
                clone.append( l.clone() )
            elif self.dbg >= ERR:
                log(ERR, '(Block - %s) cloning not implemented for: %s' \
                    % (self.__class__, l))
        return clone
    
    def show(self, with_trans=False):
        s = '%s[[[ %s ]]]\n' % (self.hierarchy*'\t', self.CallName)
        for l in self:
            s += '\t'*l.hierarchy + l.show(with_trans).replace('\n', '\n'+'\t'*l.hierarchy) + '\n'
        return s[:-1]
    
    def map(self, string=''):
        s = string
        for l in self:
            if not hasattr(l, 'Trans') or not l.Trans:
                if hasattr(l, 'parse'):
                    l.parse(s)
                else:
                    l.map(s)
                s = s[l.map_len():]
    
    # this is to retrieve full Block's dynamicity from a parsed or mapped one
    def reautomatize(self):
        for l in self:
            if hasattr(l, 'reautomatize'):
                l.reautomatize()


#------------------------------------------------------------------------------#
# testing routines 
#------------------------------------------------------------------------------#
class testTLV(Layer):
    _byte_aligned = True
    constructorList = [
        Int('T', ReprName='Tag', Type='uint8', \
            Dict={0:'Reserved', 1:'Tag1', 2:'Tag2', 5:'Tag5'}),
        Bit('F1', ReprName='Flag1', Pt=0, BitLen=1),
        Bit('F2', ReprName='Flag2', Pt=1, BitLen=2),
        Bit('res', ReprName='Reserved', Pt=0, BitLen=13),
        # length in bytes (including header, excepted Tag)
        Int('L', ReprName='Length', Type='uint8' ),
        Str('V', ReprName='Value', Pt='default value'),
        ]

    def __init__(self, **kwargs):
        Layer.__init__(self, **kwargs)
        # automating the computation of Length at runtime
        self.L.Pt = self.V
        self.L.PtFunc = lambda X: len(X)+3
        # automating the parsing of Value when calling .map(buffer)
        self.V.Len = self.L
        self.V.LenFunc = lambda X: int(X)-3

class testA(Layer):
    _byte_aligned = False
    constructorList = [
        Bit('T', ReprName='Tag', BitLen=6, Repr='hum',
            Dict={0:'Reserved', 1:'Tag1', 2:'Tag2', 5:'Tag5'}),
        Bit('F1', ReprName='Flag1', Pt=0, BitLen=4),
        Bit('F2', ReprName='Flag2', Pt=1, BitLen=2),
        # length in bytes
        Int('L', ReprName='Length', Type='uint8', Repr='hum'),
        Str('V', ReprName='Value', Pt='default value'),
        ]
    def __init__(self, **kwargs):
        Layer.__init__(self, **kwargs)
        self.L.Pt = self.V
        self.L.PtFunc = lambda X: len(X)
        self.V.Len = self.L
        self.V.LenFunc = lambda X: int(X)

class testB(Layer):
    _byte_aligned = False
    constructorList = [
        Bit('T', ReprName='Tag', BitLen=6, Repr='hum', \
            Dict={0:'Reserved', 1:'Tag1', 2:'Tag2', 5:'Tag5'}),
        Bit('F1', ReprName='Flag1', Pt=0, BitLen=4),
        Bit('F2', ReprName='Flag2', Pt=1, BitLen=2),
        # length in bits
        Int('L', Pt=0, ReprName='Length', Type='uint16', Repr='hum'),
        testA(T=1, V='super mega default value'),
        testA(T=2, V='ultra colored'),
        ]
    def __init__(self, **kwargs):
        Layer.__init__(self, **kwargs)
        self.L.PtFunc = lambda X: self[4].bit_len() + self[5].bit_len()

def test0():
    Int._endian = 'big'
    # aligned
    buf = '\x02\x22\xBB\x0Aabcdefg'
    T0 = testTLV()
    rest = T0.map_ret(buf)
    if str(T0) != buf:
        raise(Exception)
    T0.reautomatize()
    if str(T0) != buf:
        raise(Exception)

def test1():
    Int._endian = 'big'
    # non-aligned
    buf = '\xB0B\x07abcdefg'*4
    T1 = testA()
    rest = T1.map_ret(buf)
    if str(T1) != '\xb0B\x07abcdefg\xb0B\x07abcdefg\xb0B\x07abcdefg\xb0B\x07a`':
        raise(Exception)
    if str(rest)[:6] != '&6FVfp':
        raise(Exception)
    T1.reautomatize()
    if str(T1) != '\xb0B\x07abcdefg\xb0B\x07abcdefg\xb0B\x07abcdefg\xb0B\x07a`':
        raise(Exception)

def test2():
    Int._endian = 'big'
    # non-aligned
    T2 = testB()
    if str(T2) != '\x00\x10\x15\x00A\x18super mega default value\x08\x10\xd7'\
                  'V\xc7G&\x12\x066\xf6\xc6\xf7&V@':
        raise(Exception)
    buf = '\xc9\xd0B\x84v \xf2\x84\r*\xe0\xcc_\x8f\xfb\xc0\xcb\xc5\x8f\xdf\x88%\xb3'\
    'I\xc0\xe5*}]\xd7;R\x13\x87}\xa2E\xb1(\xb6\x03W`\x84&Z\x8a:eTv\xa6C\xb8\xd1'\
    '\xc48W\xa9\x82\xb1\xb89%\xc8\x18\x0b\x9e\xb7\xd6\x02\xb4\xaf\xc3J];k\x12\xf6'\
    '\xf9\x89\xb5\xdf\x92\xf8\xf0\xe7\t.0\xc2\x9c\xcd\xcb\xff1\xbd\x16K7\x1f\x95'\
    '\xbe\xbb\xda\xa8\xda\xb2\xe1x\x91U{\xf8\x06\xc1\x14\xc4%\xab\xe7S.\xac`[c'\
    '\xa8H/\x92\x8cRw\xfcpAAAA'
    rest = T2.map_ret(buf)
    if str(T2) != buf[:-4]:
        raise(Exception)
    if str(rest)[:5] != '\x04\x14\x14\x14\x10':
        raise(Exception)
    T2.reautomatize()
    if str(T2) != buf[:-4]:
        raise(Exception)

def test():
    test0()
    test1()
    test2()
#
#



# -*- coding: utf-8 -*-
"""Providing custom bootstrapper for datagen standalone app

.. module:: datagen.bootstrapper
   :platform: Unix
   :synopsis: Providing custom bootstrapper for datagen standalone app
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

import sys

PYTHON_MAJOR_VERSION = sys.version_info[0]
if PYTHON_MAJOR_VERSION == 2:
    reload(sys)
    sys.setdefaultencoding('UTF8')
    
def run_app(): 
    """Methods runs datagen standalone application
        
    Args:
       none
    
    Returns:
       void
            
    """ 
               
    from hydratk.core.masterhead import MasterHead    
    mh = MasterHead.get_head()
    mh.set_cli_cmdopt_profile('datagen')            
    mh.run_fn_hook('h_bootstrap')
    trn = mh.get_translator()  
    mh.dmsg('htk_on_debug_info', trn.msg('htk_app_exit'), mh.fromhere())                  
    sys.exit(0)


# -*- coding: UTF-8 -*-
#/**
# * Software Name : libmich 
# * Version : 0.2.3
# *
# * Copyright © 2014. Benoit Michau. ANSSI.
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License version 2 as published
# * by the Free Software Foundation. 
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# * GNU General Public License for more details. 
# *
# * You will find a copy of the terms and conditions of the GNU General Public
# * License version 2 in the "license.txt" file or
# * see http://www.gnu.org/licenses/ or write to the Free Software Foundation,
# * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
# *
# *--------------------------------------------------------
# * File Name : utils/IntEncoder.py
# * Created : 2014-06-17
# * Authors : Benoit Michau 
# *--------------------------------------------------------
#*/

###
# provides minimum encoding for signed integer
###
def minenc_int(val=0):
    '''
    for a given signed integer value, returns the minimum octet encoding.
    
    minenc_int(X) -> (Y, 'intZ')
        -2**(Y*8-1) <= Y < 2**(Y*8-1)
        'int8' -> 'intZ'
    '''
    if -128 <= val < 128:
        return (1, 'int8')
    elif -32768 <= val < 32768:
        return (2, 'int16')
    elif -8388608 <= val < 8388608:
        return (3, 'int24')
    elif -2147483648 <= val < 2147483648:
        return (4, 'int32')
    elif -549755813888 <= val < 549755813888:
        return (5, 'int40')
    elif -140737488355328 <= val < 140737488355328:
        return (6, 'int48')
    elif -36028797018963968 <= val < 36028797018963968:
        return (7, 'int56')
    elif -9223372036854775808 <= val < 9223372036854775808:
        return (8, 'int64')
    else:
        if val < 0:
            n = 9
            lim = -2**71
            while val < lim:
                n += 1
                lim <<= 8
            return (n, 'int{0}'.format(n*8))
        else:
            n = 9
            lim = 2**71
            while val >= lim:
                n += 1
                lim <<= 8
            return (n, 'int{0}'.format(n*8))

###
# provides minimum encoding for unsigned integer
###
def minenc_uint(val=0):
    '''
    for a given unsigned integer value, returns the minimum octet encoding.
    
    minenc_uint(X) -> (Y, 'uintZ')
        0 <= X < 2**(8*Y)
        'uint8' -> 'uintZ'
    '''
    if val < 0:
        return (0, 'uint0')
    elif val < 256:
        return (1, 'uint8')
    elif val < 65536:
        return (2, 'uint16')
    elif val < 16777216:
        return (3, 'uint24')
    elif val < 4294967296:
        return (4, 'uint32')
    elif val < 1099511627776:
        return (5, 'uint40')
    elif val < 281474976710656:
        return (6, 'uint48')
    elif val < 72057594037927936:
        return (7, 'uint56')
    elif val < 18446744073709551616:
        return (8, 'uint64')
    else:
        n = 9
        lim = 2**72
        while val >= lim:
            n += 1
            lim <<= 8
        return (n, 'uint{0}'.format(n*8))



# -*- coding: UTF-8 -*-
#/**
# * Software Name : libmich 
# * Version : 0.2.3
# *
# * Copyright � 2015. Benoit Michau. ANSSI.
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License version 2 as published
# * by the Free Software Foundation. 
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# * GNU General Public License for more details. 
# *
# * You will find a copy of the terms and conditions of the GNU General Public
# * License version 2 in the "license.txt" file or
# * see http://www.gnu.org/licenses/ or write to the Free Software Foundation,
# * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
# *
# *--------------------------------------------------------
# * File Name : asn1/BER.py
# * Created : 2015-01-27
# * Authors : Benoit Michau 
# *--------------------------------------------------------
#*/

# export filter
__all__ = ['BER', 'T', 'L', 'BER_TLV']

from hydratk.extensions.datagen.asn1.core.element import Element, Str, Int, Bit, Layer, show
from hydratk.extensions.datagen.asn1.core.shtr import shtr
from hydratk.extensions.datagen.asn1.utils.IntEncoder import *
#
import ASN1
from utils import *

class ASN1_BER_ENCODER(ASN1_CODEC): pass
class ASN1_BER_DECODER(ASN1_CODEC): pass

################################################################################
# For each ASN.1 object that we want to encode / decode with BER, we have:
# - tag: 1 or more bytes
#   -> encoding ASN.1 object's tag 
#   -> plus 1 bit for basic / constructed type distinction
# - length: 1 or more bytes, length of the value field in bytes
# - value: 0 or more bytes, value assigned to the ASN.1 object
#
# WARNING: several limitations exists in this implementation
# - EXPLICIT tagging, as I understand it, is not supported, except for specific
# CHOICE or OPEN types within SEQUENCE / SET
################################################################################
# few BER internal structures / naming which are useful:
# T: tag
# L: length
# V: value
# BER_TLV: standard TLV structure
################################################################################

_IntToStruct = {
    1 : 'b',
    2 : 'h',
    4 : 'i',
    8 : 'q',
    }

#------------------------------------------------------------------------------#
# BER-specific internal objects for encoding / decoding different types
#------------------------------------------------------------------------------#
BERTagClass_dict = {
    0 : TAG_UNIVERSAL,
    1 : TAG_APPLICATION,
    2 : TAG_CONTEXT_SPEC,
    3 : TAG_PRIVATE
    }
BERTagPC_dict = {
    0 : 'Primitive',
    1 : 'Constructed',
    }
class T(Layer):
    _dict = {31:'extended form'}
    _byte_aligned = True
    constructorList = [
        Bit('Class', Pt=0, BitLen=2, Repr='hum', Dict=BERTagClass_dict),
        Bit('PC', Pt=0, BitLen=1, Repr='hum', Dict=BERTagPC_dict),
        Bit('T', Pt=0, BitLen=5, Repr='hum')
        ]
    
    def __init__(self, **kwargs):
        Layer.__init__(self, **kwargs)
        if self.T() == 31:
            self.set( kwargs['T'] )
        self.T.Dict = self.Class
        self.T.DictFunc = self._get_tag_dict
    
    def _get_tag_dict(self, cla):
        if cla() == 0:
            return TAG_UNIV_VALTOTYPE
        else:
            return self._dict
    
    def set(self, T=0):
        if T <= 30:
            self.T > T
        else:
            self.T.Pt = 31
            self._val = T
            T = bin(T)[2:] # string of '100110010'
            # pad the Tag with bits 0 on the left side to be a multiple
            # of 7-bits word
            tag_len, r = len(T)//7, len(T)%7
            if r:
                T = (7-r)*'0' + T
                tag_len += 1
            # interpolate a bit 1 after each block of 7 bits
            # except fot last block which get a bit 0
            # and convert it to a string
            tag_str = ''.join(map(lambda v: chr(int(v, 2)),
                                  ['1'+T[i:i+7] for i in \
                                   range(0, (tag_len-1)*7, 7)]+['0'+T[-7:]]))
            self.append( Str('T_ext', Pt=tag_str, Repr='hex') )
    
    def get(self):
        return self()
    
    def map(self, s=''):
        if len(s) < 1:
            return ''
        if len(self.elementList) != 3:
            self.__init__()
        Layer.map(self, s[0])
        if self.T() == 31:
            s = s[1:]
            # extended form: consume the buffer until a MSBit is null
            cur = 0
            while cur < len(s)-1:
                if ord(s[cur]) & 0x80 == 0x80:
                    cur += 1
                else:
                    break
            tag_str = s[:cur+1]
            self.append( Str('T_ext', Pt=tag_str, Repr='hex') )
            # get the integral value after removing MSBit of each byte
            self._val = reduce(lambda x,y: (x<<7)+y, 
                               map(lambda c: ord(c)&0x7F, tag_str))
    
    def __call__(self):
        if not hasattr(self, '_val'):
            return self.T()
        else:
            return self._val
    
    def __repr__(self):
        t_val = self()
        if t_val in self._dict:
            t_val = '%s - %s' % (t_val, self._dict[t_val])
        return '<Tag [%s - %s]: %s>' % (self.Class.Dict[self.Class()],
                                        self.PC.Dict[self.PC()],
                                        t_val)

class L(Layer):
    _byte_aligned = True
    constructorList = [
        Bit('Form', Pt=0, BitLen=1, Repr='hum', Dict={0:'short', 1:'long'}),
        Bit('L', Pt=0, BitLen=7, Repr='hum')
        ]
    
    def __init__(self, **kwargs):
        if 'indefinite' in kwargs and kwargs['indefinite']:
            # indefinite form
            Layer.__init__(self, Form=1, L=0)
        else:
            # default short definite form
            Layer.__init__(self, **kwargs)
            if self.Form():
                # long definite form
                self.set(kwargs['L'], force_long=True)
    
    def set(self, L=0, force_long=False):
        if L <= 127 and not force_long:
            self.Form.Pt = 0
            self.L.Pt = L
        else:
            self.Form.Pt = 1
            # encode the length in the minimum number of bytes
            len_hex = hex(L)[2:]
            if len_hex[-1] == 'L':
                len_hex = len_hex[:-1]
            if len(len_hex) % 2:
                len_hex = '0'+len_hex
            self.L.Pt = len(len_hex)//2
            self.append( Bit('L_ext', Pt=L, BitLen=len(len_hex)*4, Repr='hum') )
    
    def get(self):
        return self()
    
    def map(self, s=''):
        if len(s) < 1:
            return ''
        if len(self.elementList) != 2:
            self.__init__()
        Layer.map(self, s[0])
        if self.Form() and self.L():
            self.append( Bit('L_ext', BitLen=self.L()*8, Repr='hum') )
            self[-1].map(s[1:])
    
    def __call__(self):
        if self.Form():
            if self.L() == 0:
                return -1
            else:
                return self.L_ext()
        else:
            return self.L()
    
    def __repr__(self):
        form = self.Form.Dict[self.Form()]
        if self() == -1:
            form = 'indefinite'
        return '<Len [%s]: %s>' % (form, self())

class BER_TLV(Layer):
    _LEN_MAX_INDEF = 1048576
    _LEN_FORCE_LONG = False
    #
    _byte_aligned = True
    constructorList = [
        T(),
        L(),
        Str('V', Pt='')
        ]
    
    def __init__(self, name='', **kwargs):
        Layer.__init__(self)
        if name:
            self.CallName = name
        if 'Tag' in kwargs:
            self.T.set(kwargs['Tag'])
        elif 'T' in kwargs:
            self.T.set(kwargs['T'])
        if 'Val' in kwargs:
            self.V.Pt = kwargs['Val']
            self.L.set(len(self.V), force_long=self._LEN_FORCE_LONG)
        elif 'V' in kwargs:
            self.V.Pt = kwargs['V']
            self.L.set(len(self.V), force_long=self._LEN_FORCE_LONG)
        self.V.Len = self.L
        self.V.LenFunc = lambda l: l() if l() >= 0 else self._LEN_MAX_INDEF
        # TODO: for indefinite length
        # we should better parse recursively BER_TLV struct within V
        # when of type constructed
    
    def set(self, V=''):
        if isinstance(V, BER_TLV):
            self.remove(self[2])
            self.append(V)
            self[0].PC > 1
            self[1].set(len(V), force_long=self._LEN_FORCE_LONG)
        elif isinstance(V, (tuple, list)) \
        and all([isinstance(comp, BER_TLV) for comp in V]):
            self.remove(self.V)
            for comp in V:
                self.append(comp)
            self[0].PC > 1
            self[1].set(sum(map(len, V)), force_long=self._LEN_FORCE_LONG)
        elif isinstance(V, (Element, Layer)):
            self.replace(self.V, V)
            self[1].set(len(V), force_long=self._LEN_FORCE_LONG)
        else:
            self.V.Pt = V
            self[1].set(len(V), force_long=self._LEN_FORCE_LONG)
    
    def get(self):
        return self[2]
    
    def parse(self, buf=''):
        self.__init__(self.CallName)
        self.map(buf)
        L = len(self)
        #
        # parse nested BER_TLV structures within the Value
        if self[0].PC():
            # indefinite form
            indef = self[1]() == -1
            # parsing further internal Value
            buf = str(self[2])
            inner = []
            marker = None
            while buf:
                tlv = BER_TLV()
                tlv.parse(buf)
                buf = buf[len(tlv):]
                if indef and (tlv[0]() == 0 and tlv[1]() == 0):
                    marker = tlv
                    marker.CallName = '_end_'
                    break
                else:
                    inner.append(tlv)
            if inner:
                self.remove(self[2])
                V = Layer('V')
                V.extend(inner)
                self.append(V)
            if marker:
                self.append(marker)
        # returns the length parsed
        return L

#------------------------------------------------------------------------------#
# BER encoder / decoder
#------------------------------------------------------------------------------#

class BER(ASN1.ASN1Codec):
    #
    # used by ASN1Obj types
    _name = 'BER'
    _enc_err = ASN1_BER_ENCODER
    _dec_err = ASN1_BER_DECODER
    #
    # add some costly verification when decoding buffers
    _SAFE = True
    #
    # CODEC customizations:
    # encoder customizations
    # always force setting length with the long form
    _ENC_LEN_FORCE_LONG = False
    # static value given to BOOLEAN TRUE 
    _ENC_BOOL_TRUE = 0xFF
    # force integer length for INTEGER
    _ENC_INT_LEN = None # could be any multiple of 8 bits (e.g. 64, 1024, ...)
    # force integer length for ENUMERATED
    _ENC_ENUM_LEN = None # just like _ENC_INT_LEN
    #
    # to build dictionnary for encoded / decoded ENUMERATED, CHOICE, ...
    _ENUM_BUILD_DICT = True
    #
    # libmich layers' representation (only for basic types)
    _REPR_BOOL = 'hex'
    _REPR_INT = 'hum'
    _REPR_ENUM = 'hum'
    _REPR_BIT_STR = 'hex'
    _REPR_OCT_STR = 'hex'
    _REPR_PRINT_STR = 'hum'
    
    #--------------------------------------------------------------------------#
    # encoder
    #--------------------------------------------------------------------------#
    def encode(self, obj, **kwargs):
        BER_TLV._LEN_FORCE_LONG = self._ENC_LEN_FORCE_LONG
        #
        # call the appropriate type encoder
        if obj._type == TYPE_NULL:
            self.encode_null(obj)
        elif obj._type == TYPE_BOOL:
            self.encode_bool(obj)
        elif obj._type == TYPE_INTEGER:
            self.encode_int(obj)
        elif obj._type == TYPE_ENUM:
            self.encode_enum(obj)
        elif obj._type == TYPE_BIT_STR:
            self.encode_bit_str(obj)
        elif obj._type in (TYPE_OCTET_STR, TYPE_IA5_STR, TYPE_PRINT_STR, 
                           TYPE_NUM_STR):
            self.encode_oct_str(obj)
        elif obj._type == TYPE_CHOICE:
            self.encode_choice(obj)
        elif obj._type == TYPE_SEQ:
            self.encode_seq(obj)
        elif obj._type == TYPE_SEQ_OF:
            self.encode_seq_of(obj)
        elif obj._type == TYPE_SET:
            self.encode_set(obj)
        elif obj._type == TYPE_SET_OF:
            self.encode_set_of(obj)
        elif obj._type in (TYPE_ANY, TYPE_OPEN):
            self.encode_open_type(obj)
        else:
            raise(ASN1_BER_ENCODER('%s: unsupported ASN.1 type: %s'\
                  % (obj.get_fullname(), obj._type)))
    
    def handle_tag_enc(self, obj):
        # get tag: class, primitive / constructed (pc), val
        tag = obj.get_tag()
        if tag is None:
            raise(ASN1_BER_ENCODER('%s: no tag retrieved' % obj.get_fullname()))
        mode = tag[1]
        cla, val = obj.get_tag_val()
        if obj._type in TYPE_CONSTRUCTED:
            pc = 1
        else:
            pc = 0
        #
        # 1) tag in IMPLICIT mode by default
        obj._msg[0].Class.Pt = cla
        obj._msg[0].PC.Pt = pc
        obj._msg[0].set(val)
        #
        # 2) for EXPLICIT mode, if not already UNIVERSAL, add an encapsulation
        if mode == TAG_EXPLICIT and cla != 0:
            inner = BER_TLV(obj.get_name())
            inner[0].Class.Pt = 0
            inner[0].PC.Pt = pc
            inner[0].set( TAG_UNIV_TYPETOVAL[obj._type] )
            inner.set( obj._msg.get() )
            obj._msg.set( inner )
            obj._msg[0].PC.Pt = 1
    
    #--------------------------------------------------------------------------#
    # NULL / BOOLEAN
    #--------------------------------------------------------------------------#
    def encode_null(self, obj):
        obj._msg = BER_TLV(obj.get_name(), V='')
        self.handle_tag_enc(obj)
    
    def encode_bool(self, obj):
        # obj._val = True / False
        obj._msg = BER_TLV(obj.get_name())
        obj._msg.set( chr((0, self._ENC_BOOL_TRUE)[obj._val]) )
        obj._msg[-1].Repr = self._REPR_BOOL
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # INTEGER
    #--------------------------------------------------------------------------#
    def encode_int(self, obj):
        # determine the length required to encode as a standard integer
        if isinstance(self._ENC_INT_LEN, (int, long)) \
        and self._ENC_INT_LEN % 8 == 0:
            # encoder forcing INTEGER length
            bit_len = abs(self._ENC_INT_LEN)
        else:
            # minimum byte length for INTEGER
            bit_len = 1 + len_bits(abs(obj._val))
            miss_len = bit_len % 8
            if miss_len != 0:
                bit_len += 8 - miss_len
        # encode it with libmich Int
        obj._msg = BER_TLV(obj.get_name()) 
        obj._msg.set( Int('V', Pt=obj._val, Type='int%s'%bit_len,
                          Repr=self._REPR_INT) )
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # ENUMERATION
    #--------------------------------------------------------------------------#
    def encode_enum(self, obj):
        # empty enum
        if len(obj._cont) == 0:
            obj._msg = BER_TLV(self.get_name())
            return
        # get the maximum integer value to know the length on which to encode
        if isinstance(self._ENC_ENUM_LEN, (int, long)) \
        and self._ENC_ENUM_LEN % 8 == 0:
            # encoder forcing ENUMERATED length
            bit_len = abs(self._ENC_ENUM_LEN)
        else:
            # minimum byte length for ENUMERATED
            bit_len = 1 + len_bits(max(obj._cont.values()))
            miss_len = bit_len % 8
            if miss_len != 0:
                bit_len += 8 - miss_len
        # and get the identifier integer value from the string value
        V = Int('V', Pt=obj._cont[obj._val], Type='int%s'%bit_len,
                Repr=self._REPR_ENUM)
        if self._ENUM_BUILD_DICT:
            V.Dict = dict([(i[1], i[0]) for i in obj._cont.items()])
        #
        obj._msg = BER_TLV(obj.get_name())
        obj._msg.set(V)
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # BIT STRING
    #--------------------------------------------------------------------------#
    def encode_bit_str(self, obj):
        if isinstance(obj._val, ASN1.ASN1Obj):
            # obj._val: ASN1Obj, according to CONTAINING constraint
            V = Layer('V')
            V.append(Int('pad_len', Pt=0, Type='uint8'))
            obj._val._encode()
            V.extend(obj._val._msg)
            obj._val._msg = None
        #
        else:
            # obj._val: (BE uint value, BE uint bit length)
            val_len = obj._val[1] // 8
            val_ext = obj._val[1] % 8
            if val_ext:
                val_len += 1
            V = Layer('V')
            V.append(Int('pad_len', Pt=(8-val_ext)%8, Type='uint8'))
            V.append(Bit('val', Pt=obj._val[0], BitLen=obj._val[1], 
                         Repr=self._REPR_BIT_STR))
            if val_ext > 0:
                V.append(Bit('pad', Pt=0, BitLen=8-val_ext, 
                             Repr=self._REPR_BIT_STR))
        #
        obj._msg = BER_TLV(obj.get_name())
        obj._msg.set(V)
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # OCTET STRING
    #--------------------------------------------------------------------------#
    def encode_oct_str(self, obj):
        if isinstance(obj._val, ASN1.ASN1Obj):
            # obj._val: ASN1Obj, according to CONTAINING constraint
            obj._val._encode()
            V = obj._val._msg
            obj._val._msg = None
            obj._msg = BER_TLV(obj.get_name())
            obj._msg.set(V)
        #
        else:
            # obj._val: str
            obj._msg = BER_TLV(obj.get_name(), V=obj._val)
        #
        if obj._type == TYPE_PRINT_STR:
            obj._msg[-1].Repr = self._REPR_PRINT_STR
        else:
            obj._msg[-1].Repr = self._REPR_OCT_STR
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # CHOICE
    #--------------------------------------------------------------------------#
    def encode_choice(self, obj):
        # obj._val = (choice name, type-dependent value)
        # if choice is tagged, need to encode it like a constructed one
        # otherwise, just encode the chosen component
        cho = obj._cont[obj._val[0]]
        cho._val = obj._val[1]
        cho._encode()
        tag = obj.get_tag_val()
        if tag is not None:
            obj._msg = BER_TLV(obj.get_name())
            obj._msg[0].Class.Pt = tag[0]
            obj._msg[0].PC.Pt = 1
            obj._msg[0].set(tag[1])
            obj._msg.set(cho._msg)
        else:
            cho._msg.CallName = '%s.%s' % (obj.get_name(), cho._msg.CallName)
            obj._msg = cho._msg
        cho._val = None
        cho._msg = None
    
    #--------------------------------------------------------------------------#
    # SEQUENCE
    #--------------------------------------------------------------------------#
    # values with DEFAULT and grouped extension are checked and enforced by
    # .set_val() and ._encode() methods in ASN1.ASN1Obj
    #
    # WNG: in order to provide testing facilities,
    # there is no checking on CONST_SET_REF constraint
    def encode_seq(self, obj):
        V = Layer('V')
        for name in obj._cont:
            if name in obj._val:
                comp = obj._cont[name]
                comp._val = obj._val[name]
                comp._encode()
                V.append(comp._msg)
                comp._msg = None
        #
        obj._msg = BER_TLV(obj.get_name())
        obj._msg.set(V)
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # SEQUENCE OF
    #--------------------------------------------------------------------------#
    def encode_seq_of(self, obj):
        V = Layer('V')
        for val in obj._val:
            obj._cont._val = val
            obj._cont._encode()
            V.append(obj._cont._msg)
        #
        obj._cont._msg = None
        obj._msg = BER_TLV(obj.get_name())
        obj._msg.set(V)
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # SET
    #--------------------------------------------------------------------------#
    def encode_set(self, obj):
        V = Layer('V')
        for name in obj._cont:
            if name in obj._val:
                comp = obj._cont[name]
                comp._val = obj._val[name]
                comp._encode()
                V.append(comp._msg)
                comp._msg = None
        #
        obj._msg = BER_TLV(obj.get_name())
        obj._msg.set(V)
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # SET OF
    #--------------------------------------------------------------------------#
    def encode_set_of(self, obj):
        V = Layer('V')
        for val in obj._val:
            obj._cont._val = val
            obj._cont._encode()
            V.append(obj._cont._msg)
        #
        obj._cont._msg = None
        obj._msg = BER_TLV(obj.get_name())
        obj._msg.set(V)
        self.handle_tag_enc(obj)
    
    #--------------------------------------------------------------------------#
    # OPEN TYPE
    #--------------------------------------------------------------------------#
    def encode_open_type(self, obj):
        # I did not really understand how it goes here...
        # default tag is UNIVERSAL OCTET STRING, 
        # when nothing is specified for the encapsulated type
        default_tag = (2, 4)
        #
        if isinstance(obj._cont, ASN1.ASN1Obj) and isinstance(obj._val, tuple) \
        and obj._val[0] == obj._cont._name:
            cont = obj._cont
            cont._val = obj._val[1]
            cont._encode()
            obj._msg = cont._msg
            cont._val = None
            cont._msg = None
        elif isinstance(obj._val, str):
            obj._msg = BER_TLV(obj.get_name(), V=obj._val)
            obj._msg[0].Class.Pt = default_tag[0]
            obj._msg[0].set( default_tag[1] )
        else:
            obj._msg = BER_TLV(obj.get_name(), V='')
            obj._msg[0].Class.Pt = default_tag[0]
            obj._msg[0].set( default_tag[1] )
        #
        tag = obj.get_tag_val()
        if tag is not None:
            encap = BER_TLV(obj.get_name())
            encap[0].Class.Pt = tag[0]
            encap[0].set(tag[1])
            encap.set(obj._msg)
            obj._msg = encap
    
    #--------------------------------------------------------------------------#
    # decoder
    #--------------------------------------------------------------------------#
    
    def decode(self, obj, buf, **kwargs):
        obj._msg = BER_TLV(obj.get_name())
        # this makes use of the generic (nested) BER_TLV decoder
        # whatever type obj is
        L = obj._msg.parse(buf)
        self.decode_val(obj)
        return buf[L:]
    
    def decode_val(self, obj):
        #
        # call the appropriate value type decoder     
        if obj._type == TYPE_NULL:
            return self.decode_null_val(obj)
        elif obj._type == TYPE_BOOL:
            return self.decode_bool_val(obj)
        elif obj._type == TYPE_INTEGER:
            return self.decode_int_val(obj)
        elif obj._type == TYPE_ENUM:
            return self.decode_enum_val(obj)
        elif obj._type == TYPE_BIT_STR:
            return self.decode_bit_str_val(obj)
        elif obj._type in (TYPE_OCTET_STR, TYPE_IA5_STR, TYPE_PRINT_STR,
                           TYPE_NUM_STR):
            return self.decode_oct_str_val(obj)
        elif obj._type == TYPE_CHOICE:
            return self.decode_choice_val(obj)
        elif obj._type == TYPE_SEQ:
            return self.decode_seq_val(obj)
        elif obj._type == TYPE_SEQ_OF:
            return self.decode_seq_of_val(obj)
        elif obj._type == TYPE_SET:
            return self.decode_set_val(obj)
        elif obj._type == TYPE_SET_OF:
            return self.decode_set_of_val(obj)
        elif obj._type in (TYPE_ANY, TYPE_OPEN):
            return self.decode_open_type_val(obj)
        else:
            raise(ASN1_BER_DECODER('%s: unsupported ASN.1 type'\
                  % obj.get_fullname()))
    
    def handle_tag_dec(self, obj):
        # this get the tag (IMPLICIT case) or chain of tags until the
        # UNIVERSAL one (EXPLICIT case),
        # and returns:
        #   the length of the outermost BER_TLV
        #   the innermost BER_TLV
        #
        tag = obj.get_tag()
        if tag is None:
            raise(ASN1_BER_DECODER('%s: no tag retrieved' % obj.get_fullname()))
        val, mode, cla = tag
        #
        tlv = obj._msg
        #
        if self._SAFE and tlv[0]() not in (0, val):
            raise(ASN1_BER_DECODER('%s: invalid tag provided %s compared to '\
                  'expected one %s' % (obj.get_fullname(), repr(tlv[0]), 
                                       repr(tag))))
            # TODO: we should also enforce the comparison of the Class
            # of the tag
        #
        if mode == TAG_EXPLICIT:
            # retrieve UNIVERSAL tag from the nested TLV structures
            while tlv[0].PC() and tlv[0].Class() != 0:
                tlv = tlv.get()
                if not isinstance(tlv[0], T):
                    raise(ASN1_BER_DECODER('%s: unable to retrieve UNIVERSAL '\
                      'tag' % obj.get_fullname()))
            #
            if self._SAFE and (tlv[0].Class() != 0 \
             or tlv[0]() not in (0, TAG_UNIV_TYPETOVAL[obj._type])):
                raise(ASN1_BER_DECODER('%s: invalid UNIVERSAL tag provided %s'\
                      'compared to %s' % (obj.get_fullname(), repr(tlv[0]), 
                      TAG_UNIV_TYPETOVAL[obj._type])))
            #
        return tlv
    
    #--------------------------------------------------------------------------#
    # NULL / BOOLEAN
    #--------------------------------------------------------------------------#
    def decode_null_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        obj._val = None
    
    def decode_bool_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        tlv[2].Repr = self._REPR_BOOL
        if len(tlv[2]):
            if str(tlv[2])[0] == '\0':
                obj._val = False
            else:
                obj._val = True
    
    #--------------------------------------------------------------------------#
    # INTEGER
    #--------------------------------------------------------------------------#
    
    def decode_int_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        # map to an Int, given the length of the value part
        bit_len = tlv[1]()*8
        V = Int('V', Type='int%s'%bit_len, Repr=self._REPR_INT)
        V.map(str(tlv[2]))
        tlv.set(V)
        #
        obj._val = V()
    
    #--------------------------------------------------------------------------#
    # ENUMERATION
    #--------------------------------------------------------------------------#
    
    def decode_enum_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        # map to an Int, given the length of the value part
        bit_len = tlv[1]()*8
        if bit_len:
            V = Int('V', Type='int%s'%bit_len, Repr=self._REPR_ENUM)
            V.map(str(tlv[2]))
            if self._ENUM_BUILD_TYPE:
                V.Dict = dict([(i[1], i[0]) for i in obj._cont.items()])
            tlv.set(V)
            #
            val_int = V()
            enum_int = obj._cont.values()
            if val_int in enum_int:
                obj._val = obj._cont.keys()[enum_int.index(val_int)]
            elif obj._ext is not None:
                # if extended index value is unknown,
                # the decoded value cannot be retrieved
                # WNG: we put a dummy string here as value
                obj._val = '_ext_%i' % val_int
            elif self._SAFE:
                raise(ASN1_BER_DECODER('%s: invalid ENUMERATED identifier %s' \
                      % (obj.get_fullname(), val_int)))
    
    #--------------------------------------------------------------------------#
    # BIT STRING
    #--------------------------------------------------------------------------#
    # TODO: decode to another ASN1Obj corresponding to the CONTAINING constraint
    
    def decode_bit_str_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        # obj._val: (BE uint value, BE uint bit length)
        # map to a Layer including the pad_len initial octet
        v_str = str(tlv[2])
        if len(v_str) > 1:
            pad_len = ord(v_str[0])
            bit_len = 8*(len(v_str)-1)-pad_len
            V = Layer('V')
            V.append(Int('pad_len'), Type='uint8')
            V.append(Bit('val', BitLen=bit_len, Repr=self._REPR_BIT_STR))
            if pad_len > 0:
                V.append(Bit('pad', BitLen=pad_len, Repr=self._REPR_BIT_STR))
            V.map(v_str)
            tlv.set(V)
            #
            obj._val = (V[1](), bit_len)
        else:
            obj._val = (0, 0)
    
    
    #--------------------------------------------------------------------------#
    # OCTET STRING
    #--------------------------------------------------------------------------#
    # TODO: decode to another ASN1Obj corresponding to the CONTAINING constraint
    
    def decode_oct_str_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        if obj._type == TYPE_PRINT_STR:
            tlv[2].Repr = self._REPR_PRINT_STR
        else:
            tlv[2].Repr = self._REPR_OCT_STR
        obj._val = str(tlv[2])
    
    #--------------------------------------------------------------------------#
    # CHOICE
    #--------------------------------------------------------------------------#
    # TODO: decode unknown extension according to their UNIVERSAL tag,
    # when provided by the encoder
    #    
    def decode_choice_val(self, obj):
        # obj._val = (name, value)
        # check if CHOICE is tagged by the application
        
        tag = obj.get_tag()
        if tag is not None:
            # if CHOICE is tagged, need to get its tag first
            if self._SAFE and obj._msg[0]() != tag[0]:
                raise(ASN1_BER_DECODER('%s: invalid CHOICE tag provided %s ' \
                      'compared to expected one %s' \
                      % (obj.get_fullname(), repr(obj._msg[0]), tag)))
            tlv = obj._msg.get()[0] # hydratk fix
        else:
            # otherwise, tag is directly corresponding to the chosen object
            tlv = obj._msg
        tag = (tlv[0].Class(), tlv[0]())
        #
        if tag in obj._cont_tags:
            # select the chosen object according to the tag
            cho_names = obj._cont_tags[tag]
            cho = obj._cont[cho_names[0]]
            if len(cho_names) > 1:
                # chain of untagged CHOICE
                for name in cho_names[1:]:
                    cho = cho._cont[name]
            #
            obj._val = nest(cho_names,
                            self._decode_comp_val(obj, cho, tlv))
        #
        elif obj._ext is not None:
            # extended unknown object
            obj._val = ('_ext_%i' % tag[1], str(tlv[2]))
            # TODO, if we have an UNIVERSAL tag, we can decode the passed value
        #
        elif self._SAFE:
            raise(ASN1_BER_DECODER('%s: invalid CHOICE tag %s' \
                  % (obj.get_fullname(), tag)))
    
    def _decode_comp_val(self, obj, comp, tlv):
        #print('decoding value for component %s of %s' \
        #      % (comp._name, obj.get_fullname()))
        if comp._type in (TYPE_OPEN, TYPE_ANY):
            const = comp.get_const_ref()
            if const:
                done, ref = self._get_open_ref(obj, comp, const)
                if done:
                    #log('CONST_SET_REF, ref: %s' % ref._name)
                    comp._cont = ref
        #
        comp = comp.clone_light()
        comp._msg = tlv
        tlv.CallName = comp.get_name()
        self.decode_val(comp)
        val = comp._val
        comp._val = None
        #comp._msg = None
        #print('decoded %s' % comp._name)
        return val
    
    def _get_open_ref(self, obj, comp, const):
        if not const['at']:
            raise(ASN1_OBJ('%s: invalid SET_REF constraint'\
                  % obj.get_fullname()))
        # get the value referred by @ identifier from obj
        if const['at'] not in obj._val:
            # this would mean the @ identifier field is placed 
            # after the OPEN TYPE to decode
            # -> terrible protocol design !
            #raise(ASN1_OBJ('%s: not able to retrieve value for @ '\
            #      'component %s' % (obj.get_fullname(), const['at'])))
            return (False, None)
        #
        at_val = obj._val[const['at']]
        #log('_get_open_ref, at: %s, at_val: %s' % (const['at'], at_val))
        try:
            ref = const['ref'](const['at'], at_val)
        except:
            # invalid value passed against object info value set
            return (False, None)
        #
        comp_typename = comp.get_typename()
        #log('_get_open_ref, comp_typename: %s' % comp_typename)
        if comp_typename not in ref:
            raise(ASN1_OBJ('%s: not able to retrieve %s within '\
                  'object info set' % (obj.get_fullname(), comp_typename)))
        return (True, ref[comp_typename].clone_light()) 
    
    #--------------------------------------------------------------------------#
    # SEQUENCE
    #--------------------------------------------------------------------------#
    # TODO: 
    # - untagged OPEN / ANY types are not handled, which will certainly raise
    # if present in the encoded SEQUENCE
    # - value of unknown extended components are not handled 
    def decode_seq_val(self, obj):
        val_aux = obj._val # hydratk fix
        tlv = self.handle_tag_dec(obj)
        compts = tlv.get()
        # compts must be Layer containing BER_TLV layers
        if self._SAFE \
        and not all([isinstance(comp, BER_TLV) for comp in compts]):
            raise(ASN1_BER_DECODER('%s: invalid SEQUENCE encoded content'\
                  % obj.get_fullname()))
        #
        # cursors for going over the components and _cont_tags list
        cur_compts = 0
        cur_tag = 0
        obj._val = dict()
        # going over all components of the compiled SEQUENCE
        for name in obj._cont:
            if cur_compts >= len(compts.elementList):
                break
            tlv_inner = compts[cur_compts]            
            #print('cur: %s, component: %s' % (cur_compts, repr(tlv_inner)))
            tag = (tlv_inner[0].Class(), tlv_inner[0]())
            comp = obj._cont[name]
            comp_tag = comp.get_tag_val()
            #print('expected name / tag: %s / %s' % (name, repr(comp_tag)))
            #
            if comp_tag is None:
                # untagged component: CHOICE / OPEN / ANY type
                # TODO: handle OPEN / ANY types
                #
                # get all possible tags within the CHOICE
                if comp._type == TYPE_CHOICE:
                    tags = [t[0] for t in obj._cont_tags[cur_tag:] \
                            if t[1][0] == name]
                    if tag in tags:
                        obj._val[name] = self._decode_comp_val(obj,
                                                               comp,
                                                               tlv_inner)
                        tlv_inner.CallName = '.'.join(obj._cont_tags[cur_tag:]\
                                                     [tags.index(tag)][1])
                        cur_compts += 1
                    cur_tag += len(tags)
                else:
                    # OPEN TYPE
                    obj._val[name] = self._decode_comp_val(obj,
                                                           comp,
                                                           tlv_inner)
                    cur_compts += 1
                    cur_tag += 1
            else:
                if comp_tag == tag:
                    # tagged component, including CHOICE / OPEN / ANY type
                    obj._val[name] = self._decode_comp_val(obj,
                                                           comp,
                                                           tlv_inner)
                    cur_compts += 1
                cur_tag += 1
        # hydratk fix
        if (len(obj._val.keys()) == 0):
            obj._val = val_aux
        #
        for name in [n for n in obj._root_comp if n not in obj._root_opt]:
            if name not in obj._val:                
                # mandatory component was not present in the decoded components
                raise(ASN1_BER_DECODER('%s: missing mandatory component %s' \
                      % (obj.get_fullname(), name)))
    
    #--------------------------------------------------------------------------#
    # SEQUENCE OF
    #--------------------------------------------------------------------------#
    def decode_seq_of_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        compts = tlv.get()
        if self._SAFE \
        and not all([isinstance(comp, BER_TLV) for comp in compts]):
            raise(ASN1_BER_DECODER('%s: invalid SEQUENCE OF encoded content'\
                  % obj.get_fullname()))
        #
        obj._val = []
        for tlv in compts:
            obj._val.append( self._decode_comp_val(obj, obj._cont, tlv) )
    
    #--------------------------------------------------------------------------#
    # SET
    #--------------------------------------------------------------------------#
    def decode_set_val(self, obj):
        tlv = self.handle_tag_dec(obj)
        compts = tlv.get()
        # compts must be Layer containing BER_TLV layers
        if self._SAFE \
        and not all([isinstance(comp, BER_TLV) for comp in compts]):
            raise(ASN1_BER_DECODER('%s: invalid SET encoded content'\
                  % obj.get_fullname()))
        #
        # remainder for components
        tags = list(obj._cont_tags)
        obj._val = dict()
        # going over all components of the decoded SET
        for tlv in compts:
            tag = (tlv[0].Class(), tlv[0]())
            if tag in obj._cont_tags:
                # known component
                if self._SAFE and tag not in tags:
                    raise(ASN1_BER_DECODER('%s: multiple SET component %s' \
                          % (obj.get_fullname(), obj._cont_tags[tag][0])))
                name = obj._cont_tags[tag][0]
                comp = obj._cont[name]
                obj._val[name] = self._decode_comp_val(obj, comp, tlv)
                tags.remove(tag)
        #
        for name in [n for n in obj._root_comp if n not in obj._root_opt]:
            if name not in obj._val:
                # mandatory component was not present in the decoded "compts"
                raise(ASN1_BER_DECODER('%s: missing mandatory component %s' \
                      % (obj.get_fullname(), name)))
    
    #--------------------------------------------------------------------------#
    # SEQUENCE OF
    #--------------------------------------------------------------------------#
    def decode_set_of(self, obj):
        tlv = self.handle_tag_dec(obj)
        compts = tlv.get()
        if self._SAFE \
        and not all([isinstance(comp, BER_TLV) for comp in compts]):
            raise(ASN1_BER_DECODER('%s: invalid SET OF encoded content'\
                  % obj.get_fullname()))
        #
        obj._val = []
        for tlv in compts:
            obj._val.append( self._decode_comp_val(obj, obj._cont, tlv) )
    
    #--------------------------------------------------------------------------#
    # OPEN TYPE
    #--------------------------------------------------------------------------#
    # TODO: decode unknown OPEN TYPE according to their UNIVERSAL tag,
    # when provided by the encoder
    #
    def decode_open_type_val(self, obj):
        # check if OPEN / ANY type is tagged by the application
        tag = obj.get_tag()
        if tag is not None:
            # if tagged, need to get its tag first
            if self._SAFE and obj._msg[0]() != tag[0]:
                raise(ASN1_BER_DECODER('%s: invalid OPEN TYPE tag provided %s ' \
                      'compared to expected one %s' \
                      % (obj.get_fullname(), repr(obj._msg[0]), tag)))
            tlv = obj._msg.get()
        else:
            # otherwise, tag is directly corresponding to the chosen object
            tlv = obj._msg
        #
        if isinstance(obj._cont, ASN1.ASN1Obj):
            obj._val = (obj._cont._name,
                        self._decode_comp_val(obj, obj._cont, tlv))
        else:
            obj._val = str(tlv[2])
    


# -*- coding: utf-8 -*-
"""Module for ASN.1 codec

.. module:: datagen.asn1codec
   :platform: Unix
   :synopsis: Module for ASN.1 codec
              Libraries for ASN.1 were taken from https://github.com/mitshell/libmich
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

"""
Events:
-------
asn1_before_import_spec
asn1_after_import_spec
asn1_before_encode
asn1_after_encode
asn1_before_decode
asn1_after_decode

"""

from hydratk.core.masterhead import MasterHead
from hydratk.core import event
from os import path
from simplejson import loads, dumps
from binascii import hexlify
from sys import version_info

if (version_info[0] == 2 and version_info[1] == 7):
    from asn1.asn1.processor import process_modules
    from asn1.asn1.ASN1 import ASN1Obj
    from asn1.asn1.BER import BER
    from collections import OrderedDict

class ASN1Codec(object):
    """Class ASN1Codec
    """
    
    _mh = None
    _path = None
    _spec = None
    _elements = None
    
    def __init__(self):
        """Class constructor
        
        Called when object is initialized
        
        Args:   
           none         
                
        """  
        
        if (not (version_info[0] == 2 and version_info[1] == 7)):
            raise NotImplementedError('ASN.1 codec is not supported for Python 2.6 and 3.x due to external library libmich')               
        
        self._mh = MasterHead.get_head()
    
    @property
    def path(self):
        """ path property getter """
        
        return self._path    
    
    @property
    def spec(self):
        """ spec property getter """
        
        return self._spec
    
    @property
    def elements(self):
        """ elements property getter """
        
        return self._elements
    
    def __str__(self):
        """Method overrides __str__
        
        Args:  
           none          
           
        Returns:
           str
                
        """         
        
        return self._spec._text_def 
    
    def import_spec(self, filename):
        """Method imports specification
        
        Args:
            filename (str): filename
            
        Returns:
            bool: result
        
        Raises:
            event: asn1_before_import_spec
            event: asn1_after_import_spec   
                
        """     
    
        try:
    
            self._mh.dmsg('htk_on_debug_info', self._mh._trn.msg('datagen_asn1_import_spec', filename), self._mh.fromhere()) 
            ev = event.Event('asn1_before_import_spec', filename)
            if (self._mh.fire_event(ev) > 0):
                filename = ev.argv(0)
                
            if (ev.will_run_default()):        
                if (path.exists(filename)):
                    with open(filename, 'r') as f:                                         
                        self._spec = process_modules(f.read())  
                        self._elements = self._spec[0]['TYPE']._dict
                        ASN1Obj._SAFE = True
                        ASN1Obj._RET_STRUCT = True
                        ASN1Obj.CODEC = BER    
                        self._path = path.abspath(filename)                                            
                else:
                    raise ValueError('File {0} not found'.format(filename))      
                
            self._mh.dmsg('htk_on_debug_info', self._mh._trn.msg('datagen_asn1_spec_imported'), self._mh.fromhere())   
            ev = event.Event('asn1_after_import_spec')
            self._mh.fire_event(ev)
                    
            return True                             
               
        except (Exception, ValueError) as ex:
            self._mh.dmsg('htk_on_error', 'error: {0}'.format(ex), self._mh.fromhere())
            return False
        
    def encode(self, infile, element, outfile=None):
        """Method encodes json file to binary 
        
        Args:
            infile (str): input filename
            element (str): element name
            outfile (str): output filename, default infile with suffix bin
            
        Returns: 
            bool: result
        
        Raises: 
            event: asn1_before_encode
            event: asn1_after_encode  
                
        """         
        
        try:
            self._mh.dmsg('htk_on_debug_info', self._mh._trn.msg('datagen_asn1_encode', infile), self._mh.fromhere()) 
            
            if (self._spec == None):
                raise ValueError('Specification is not imported yet')             
            
            ev = event.Event('asn1_before_encode', infile, element, outfile)
            if (self._mh.fire_event(ev) > 0):
                infile = ev.argv(0)
                element = ev.argv(1)
                outfile = ev.argv(2)        
        
            if (ev.will_run_default()):
                if (element not in self._elements):
                    raise ValueError('Invalid element {0}'.format(element)) 
                
                if (path.exists(infile)):         
                    with open(infile, 'r') as f: 
                        objects = loads(f.read())
                        if (objects.__class__.__name__ == 'list'):
                            input = []
                            for record in objects:
                                input.append(self._update_datatypes(record))
                        else:                                     
                            input = [self._update_datatypes(objects)]            
                    outfile = path.abspath(infile).split('.')[0]+'.bin' if (outfile == None) else outfile               
                    with open(outfile, 'wb') as f:
                        for record in input:                      
                            output = self._elements[element].encode(record)
                            f.write(str(output()))                                                         
                else:
                    raise ValueError('File {0} not found'.format(infile))
                
            self._mh.dmsg('htk_on_debug_info', self._mh._trn.msg('datagen_asn1_encoded', outfile), self._mh.fromhere())   
            ev = event.Event('asn1_after_encode')
            self._mh.fire_event(ev)
                    
            return True                 
             
        except (Exception, ValueError) as ex:
            self._mh.dmsg('htk_on_error', 'error: {0}'.format(ex), self._mh.fromhere())   
            return False     
        
    def decode(self, infile, element, outfile=None):
        """Method decodes binary file to json
        
        Args:
            infile (str): input filename
            element (str): element name
            outfile (str): output filename, default infile with suffix json
            
        Returns:
            bool: result
        
        Raises:
            event: asn1_before_decode
            event: asn1_after_decode     
                
        """         
        
        try:
        
            self._mh.dmsg('htk_on_debug_info', self._mh._trn.msg('datagen_asn1_decode', infile), self._mh.fromhere()) 
            
            if (self._spec == None):
                raise ValueError('Specification is not imported yet')             
            
            ev = event.Event('asn1_before_decode', infile, element, outfile)
            if (self._mh.fire_event(ev) > 0):
                infile = ev.argv(0)
                element = ev.argv(1)
                outfile = ev.argv(2)           
        
            if (ev.will_run_default()): 
                if (element not in self._elements):
                    raise ValueError('Invalid element {0}'.format(element))                
                
                if (path.exists(infile)):         
                    with open(infile, 'rb') as f:                                         
                        input = f.read()
                
                    outfile = path.abspath(infile).split('.')[0]+'.json' if (outfile == None) else outfile     
                    with open(outfile, 'w') as f:                      
                        records = self._split_records(input)

                        if (len(records) > 1):
                            output = []
                            for record in records:                                
                                self._elements[element].decode(record)
                                output.append(self._create_dict(self._elements[element]))
                        else:                                    
                            self._elements[element].decode(records[0])
                            output = self._create_dict(self._elements[element])
                        f.write(dumps(output, indent=4))                                                      
                else:
                    raise ValueError('File {0} not found'.format(infile)) 
            
            self._mh.dmsg('htk_on_debug_info', self._mh._trn.msg('datagen_asn1_decoded', outfile), self._mh.fromhere())   
            ev = event.Event('asn1_after_decode')
            self._mh.fire_event(ev)
                    
            return True              

        except (Exception, ValueError) as ex:
            self._mh.dmsg('htk_on_error', 'error: {0}'.format(ex), self._mh.fromhere())  
            return False             
        
    def _update_datatypes(self, obj):
        """Method updates datatypes
        
        It is used in encoder, some datatypes can't be encoded
        The object is traversed recursively
        
        Args:
            obj (json): particular json object  
            
        Returns:
            json: json object
                
        """           
        
        classname = obj.__class__.__name__
        if (classname == 'dict'):
            for key, value in obj.items():
                obj[key] = self._update_datatypes(obj[key])
        elif (classname == 'list'):
            items = []
            for item in obj:
                items.append(self._update_datatypes(item))    
            obj = items           
        elif (classname == 'long'):
            obj = int(obj) 
        elif (classname == 'unicode'):
            obj = str(obj)
           
        return obj  
    
    def _split_records(self, input):
        """Method splits records from input file
        
        Args:
            input (str): file contents
            
        Returns:
            list: list of records
                
        """           
        
        records = []
        input = bytearray(input)
        i = 0
        idx = None
        length = len(input)
        while (i < length):
            len_byte = input[i+1]
            if (len_byte < 128):
                idx = i+len_byte+2
            elif (len_byte == 128):
                idx = length
            else:
                size_bytes = len_byte-128
                rec_size = int(hexlify(input[i+2 : i+2+size_bytes]), 16)
                idx = i+2+size_bytes+rec_size
                
            records.append(str(input[i:idx]))
            i = idx
         
        return records
    
    def _create_dict(self, obj, val=None, output=None):
        """Method creates dictionary according to spec
        
        It is used in decoder, objects are ordered as specified
        The object is traversed recursively
        
        Args:
            obj (obj): particular ASN1 object
            val (obj): element value
            output (dict): particular output
            
        Returns:
            dict: ordered dictionary object 
                
        """          
        
        if (val == None):
            val = obj._val 
        if (output == None):
            output = OrderedDict()   
            
        if (obj._type in ('SEQUENCE', 'SET')):
            for key in obj._cont._index:
                if (key in val):           
                    output[key] = OrderedDict()
                    output[key] = self._create_dict(obj._cont._dict[key], val[key], output[key])
        elif (obj._type == 'SEQUENCE OF'):
            output = []
            for item in val:
                output.append(self._create_dict(obj._cont, item, None))                
        elif (obj._type == 'CHOICE'):
            key = val[0]
            output[key] = OrderedDict()
            output[key] = self._create_dict(obj._cont._dict[key], val[1], output[key])
        else:
            if (val.__class__.__name__ == 'str' and '\\x' in val.__repr__()):
                val = val.encode('hex')                
            output = val
            
        return output  


# -*- coding: utf-8 -*-

"""This code is a part of Hydra Toolkit

.. module:: hydratk.extensions.datagen.translation.en.help
   :platform: Unix
   :synopsis: English language translation for Datagen extension help generator
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""
language = {
  'name' : 'English',
  'ISO-639-1' : 'en'
} 

''' Datagen Commands '''
help_cmd = {
    'gen-asn1' : 'encode text file, decode binary file according to ASN.1 specification',
    'gen-json' : 'generate sample json file according to JSON specification',
    'gen-xml'  : 'generate sample xml file according to WSDL/XSD specification',
    
    #standalone with option profile datagen 
    'asn1' : 'encode text file, decode binary file according to ASN.1 specification',
    'json' : 'generate sample json file according to JSON specification',
    'xml'  : 'generate sample xml file according to WSDL/XSD specification'
}

''' Datagen Options '''
help_opt = {             
   'gen-spec'     : { '{h}--gen-spec <path>{e}' : { 'description' : 'specification filename', 'commands' : ('gen-asn1', 'gen-json', 'gen-xml')}},
   'gen-input'    : { '{h}--gen-input <path>{e}' : { 'description' : 'input filename', 'commands' : ('gen-asn1')}},
   'gen-output'   : { '{h}[--gen-output <path>]{e}' : { 'description' : 'output filename, default input filename with changed suffix or sample.json, sample.xml', 'commands' : ('gen-asn1', 'gen-json', 'gen-xml')}},
   'gen-action'   : { '{h}--gen-action encode|decode{e}' : { 'description' : 'action', 'commands' : ('gen-asn1')}},
   'gen-element'  : { '{h}--gen-element <title>{e}' : { 'description' : 'element title from specification', 'commands' : ('gen-asn1', 'gen-xml')}},
   'gen-envelope' : { '{h}[--gen-envelope]{e}' : { 'description' : 'generate including SOAP envelope', 'commands' : ('gen-xml')}},
   
   #standalone with option profile datagen  
   'spec'     : { '{h}--spec <path>{e}' : { 'description' : 'specification filename', 'commands' : ('asn1', 'json', 'xml')}},
   'input'    : { '{h}--input <path>{e}' : { 'description' : 'input filename', 'commands' : ('asn1')}},
   'output'   : { '{h}[--output <path>]{e}' : { 'description' : 'output filename, default input filename with changed suffix or sample.json, sample.xml', 'commands' : ('asn1', 'json', 'xml')}},
   'action'   : { '{h}--action encode|decode{e}' : { 'description' : 'action', 'commands' : ('asn1')}},
   'element'  : { '{h}--element <title>{e}' : { 'description' : 'element title from specification', 'commands' : ('asn1', 'xml')}},
   'envelope' : { '{h}[--envelope]{e}' : { 'description' : 'generate including SOAP envelope', 'commands' : ('xml')}}                  
}

